# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import datetime
import logging
from collections.abc import Callable, MutableMapping
from typing import Any, cast

from bumble import avc, avctp, avdtp, avrcp, crypto, rfcomm, sdp
from bumble.att import ATT_CID, ATT_PDU
from bumble.colors import color
from bumble.core import name_or_number
from bumble.hci import (
    HCI_ACL_DATA_PACKET,
    HCI_DISCONNECTION_COMPLETE_EVENT,
    HCI_EVENT_PACKET,
    Address,
    HCI_AclDataPacket,
    HCI_AclDataPacketAssembler,
    HCI_Disconnection_Complete_Event,
    HCI_Event,
    HCI_Packet,
)
from bumble.l2cap import (
    L2CAP_LE_SIGNALING_CID,
    L2CAP_PDU,
    L2CAP_SIGNALING_CID,
    CommandCode,
    L2CAP_Connection_Request,
    L2CAP_Connection_Response,
    L2CAP_Control_Frame,
)
from bumble.smp import SMP_CID, SMP_Command

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
PSM_NAMES = {
    rfcomm.RFCOMM_PSM: 'RFCOMM',
    sdp.SDP_PSM: 'SDP',
    avdtp.AVDTP_PSM: 'AVDTP',
    avctp.AVCTP_PSM: 'AVCTP',
    # TODO: add more PSM values
}

AVCTP_PID_NAMES = {avrcp.AVRCP_PID: 'AVRCP'}


# -----------------------------------------------------------------------------
class PacketTracer:
    class AclStream:
        psms: MutableMapping[int, int]
        peer: PacketTracer.AclStream | None
        avdtp_assemblers: MutableMapping[int, avdtp.MessageAssembler]
        avctp_assemblers: MutableMapping[int, avctp.MessageAssembler]

        def __init__(self, analyzer: PacketTracer.Analyzer) -> None:
            self.analyzer = analyzer
            self.packet_assembler = HCI_AclDataPacketAssembler(self.on_acl_pdu)
            self.avdtp_assemblers = {}  # AVDTP assemblers, by source_cid
            self.avctp_assemblers = {}  # AVCTP assemblers, by source_cid
            self.psms = {}  # PSM, by source_cid
            self.peer = None

        # pylint: disable=too-many-nested-blocks
        def on_acl_pdu(self, pdu: bytes) -> None:
            l2cap_pdu = L2CAP_PDU.from_bytes(pdu)
            self.analyzer.emit(l2cap_pdu)

            if l2cap_pdu.cid == ATT_CID:
                att_pdu = ATT_PDU.from_bytes(l2cap_pdu.payload)
                self.analyzer.emit(att_pdu)
            elif l2cap_pdu.cid == SMP_CID:
                smp_command = SMP_Command.from_bytes(l2cap_pdu.payload)
                self.analyzer.emit(smp_command)
            elif l2cap_pdu.cid in (L2CAP_SIGNALING_CID, L2CAP_LE_SIGNALING_CID):
                control_frame = L2CAP_Control_Frame.from_bytes(l2cap_pdu.payload)
                self.analyzer.emit(control_frame)

                # Check if this signals a new channel
                if control_frame.code == CommandCode.L2CAP_CONNECTION_REQUEST:
                    connection_request = cast(L2CAP_Connection_Request, control_frame)
                    self.psms[connection_request.source_cid] = connection_request.psm
                elif control_frame.code == CommandCode.L2CAP_CONNECTION_RESPONSE:
                    connection_response = cast(L2CAP_Connection_Response, control_frame)
                    if (
                        connection_response.result
                        == L2CAP_Connection_Response.Result.CONNECTION_SUCCESSFUL
                    ):
                        if self.peer and (
                            psm := self.peer.psms.get(connection_response.source_cid)
                        ):
                            # Found a pending connection
                            self.psms[connection_response.destination_cid] = psm

                            # For AVDTP connections, create a packet assembler for
                            # each direction
                            if psm == avdtp.AVDTP_PSM:
                                self.avdtp_assemblers[
                                    connection_response.source_cid
                                ] = avdtp.MessageAssembler(self.on_avdtp_message)
                                self.peer.avdtp_assemblers[
                                    connection_response.destination_cid
                                ] = avdtp.MessageAssembler(self.peer.on_avdtp_message)
                            elif psm == avctp.AVCTP_PSM:
                                self.avctp_assemblers[
                                    connection_response.source_cid
                                ] = avctp.MessageAssembler(self.on_avctp_message)
                                self.peer.avctp_assemblers[
                                    connection_response.destination_cid
                                ] = avctp.MessageAssembler(self.peer.on_avctp_message)
            else:
                # Try to find the PSM associated with this PDU
                if self.peer and (psm := self.peer.psms.get(l2cap_pdu.cid)):
                    if psm == sdp.SDP_PSM:
                        sdp_pdu = sdp.SDP_PDU.from_bytes(l2cap_pdu.payload)
                        self.analyzer.emit(sdp_pdu)
                    elif psm == rfcomm.RFCOMM_PSM:
                        rfcomm_frame = rfcomm.RFCOMM_Frame.from_bytes(l2cap_pdu.payload)
                        self.analyzer.emit(rfcomm_frame)
                    elif psm == avdtp.AVDTP_PSM:
                        self.analyzer.emit(
                            f'{color("L2CAP", "green")} [CID={l2cap_pdu.cid}, '
                            f'PSM=AVDTP]: {l2cap_pdu.payload.hex()}'
                        )
                        if avdtp_assembler := self.avdtp_assemblers.get(l2cap_pdu.cid):
                            avdtp_assembler.on_pdu(l2cap_pdu.payload)
                    elif psm == avctp.AVCTP_PSM:
                        self.analyzer.emit(
                            f'{color("L2CAP", "green")} [CID={l2cap_pdu.cid}, '
                            f'PSM=AVCTP]: {l2cap_pdu.payload.hex()}'
                        )
                        if avctp_assembler := self.avctp_assemblers.get(l2cap_pdu.cid):
                            avctp_assembler.on_pdu(l2cap_pdu.payload)
                    else:
                        psm_string = name_or_number(PSM_NAMES, psm)
                        self.analyzer.emit(
                            f'{color("L2CAP", "green")} [CID={l2cap_pdu.cid}, '
                            f'PSM={psm_string}]: {l2cap_pdu.payload.hex()}'
                        )
                else:
                    self.analyzer.emit(l2cap_pdu)

        def on_avdtp_message(
            self, transaction_label: int, message: avdtp.Message
        ) -> None:
            self.analyzer.emit(
                f'{color("AVDTP", "green")} [{transaction_label}] {message}'
            )

        def on_avctp_message(
            self,
            transaction_label: int,
            is_command: bool,
            ipid: bool,
            pid: int,
            payload: bytes,
        ):
            if pid == avrcp.AVRCP_PID:
                avc_frame = avc.Frame.from_bytes(payload)
                details = str(avc_frame)
            else:
                details = payload.hex()

            c_r = 'Command' if is_command else 'Response'
            self.analyzer.emit(
                f'{color("AVCTP", "green")} '
                f'{c_r}[{transaction_label}][{name_or_number(AVCTP_PID_NAMES, pid)}] '
                f'{"#" if ipid else ""}'
                f'{details}'
            )

        def feed_packet(self, packet: HCI_AclDataPacket) -> None:
            self.packet_assembler.feed_packet(packet)

    class Analyzer:
        acl_streams: MutableMapping[int, PacketTracer.AclStream]
        peer: PacketTracer.Analyzer

        def __init__(self, label: str, emit_message: Callable[..., None]) -> None:
            self.label = label
            self.emit_message = emit_message
            self.acl_streams = {}  # ACL streams, by connection handle
            self.packet_timestamp: datetime.datetime | None = None

        def start_acl_stream(self, connection_handle: int) -> PacketTracer.AclStream:
            logger.info(
                f'[{self.label}] +++ Creating ACL stream for connection '
                f'0x{connection_handle:04X}'
            )
            stream = PacketTracer.AclStream(self)
            self.acl_streams[connection_handle] = stream

            # Associate with a peer stream if we can
            if peer_stream := self.peer.acl_streams.get(connection_handle):
                stream.peer = peer_stream
                peer_stream.peer = stream

            return stream

        def end_acl_stream(self, connection_handle: int) -> None:
            if connection_handle in self.acl_streams:
                logger.info(
                    f'[{self.label}] --- Removing ACL stream for connection '
                    f'0x{connection_handle:04X}'
                )
                del self.acl_streams[connection_handle]

                # Let the other forwarder know so it can cleanup its stream as well
                self.peer.end_acl_stream(connection_handle)

        def on_packet(
            self, timestamp: datetime.datetime | None, packet: HCI_Packet
        ) -> None:
            self.packet_timestamp = timestamp
            self.emit(packet)

            if packet.hci_packet_type == HCI_ACL_DATA_PACKET:
                acl_packet = cast(HCI_AclDataPacket, packet)
                # Look for an existing stream for this handle, create one if it is the
                # first ACL packet for that connection handle
                if (
                    stream := self.acl_streams.get(acl_packet.connection_handle)
                ) is None:
                    stream = self.start_acl_stream(acl_packet.connection_handle)
                stream.feed_packet(acl_packet)
            elif packet.hci_packet_type == HCI_EVENT_PACKET:
                event_packet = cast(HCI_Event, packet)
                if event_packet.event_code == HCI_DISCONNECTION_COMPLETE_EVENT:
                    self.end_acl_stream(
                        cast(HCI_Disconnection_Complete_Event, packet).connection_handle
                    )

        def emit(self, message: Any) -> None:
            if self.packet_timestamp:
                prefix = f"[{self.packet_timestamp.strftime('%Y-%m-%d %H:%M:%S.%f')}]"
            else:
                prefix = ""
            self.emit_message(f'{prefix}[{self.label}] {message}')

    def trace(
        self,
        packet: HCI_Packet,
        direction: int = 0,
        timestamp: datetime.datetime | None = None,
    ) -> None:
        if direction == 0:
            self.host_to_controller_analyzer.on_packet(timestamp, packet)
        else:
            self.controller_to_host_analyzer.on_packet(timestamp, packet)

    def __init__(
        self,
        host_to_controller_label: str = color('HOST->CONTROLLER', 'blue'),
        controller_to_host_label: str = color('CONTROLLER->HOST', 'cyan'),
        emit_message: Callable[..., None] = logger.info,
    ) -> None:
        self.host_to_controller_analyzer = PacketTracer.Analyzer(
            host_to_controller_label, emit_message
        )
        self.controller_to_host_analyzer = PacketTracer.Analyzer(
            controller_to_host_label, emit_message
        )
        self.host_to_controller_analyzer.peer = self.controller_to_host_analyzer
        self.controller_to_host_analyzer.peer = self.host_to_controller_analyzer


def generate_irk() -> bytes:
    return crypto.r()


def verify_rpa_with_irk(rpa: Address, irk: bytes) -> bool:
    rpa_bytes = bytes(rpa)
    prand_given = rpa_bytes[3:]
    hash_given = rpa_bytes[:3]
    hash_local = crypto.ah(irk, prand_given)
    return hash_local[:3] == hash_given



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import pytest
import pytest_asyncio

from bumble import device
from bumble.profiles.gmap import (
    BgrFeatures,
    BgsFeatures,
    GamingAudioService,
    GamingAudioServiceProxy,
    GmapRole,
    UggFeatures,
    UgtFeatures,
)

from .test_utils import TwoDevices

# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------
gmas_service = GamingAudioService(
    gmap_role=GmapRole.UNICAST_GAME_GATEWAY
    | GmapRole.UNICAST_GAME_TERMINAL
    | GmapRole.BROADCAST_GAME_RECEIVER
    | GmapRole.BROADCAST_GAME_SENDER,
    ugg_features=UggFeatures.UGG_MULTISINK,
    ugt_features=UgtFeatures.UGT_SOURCE,
    bgr_features=BgrFeatures.BGR_MULTISINK,
    bgs_features=BgsFeatures.BGS_96_KBPS,
)


@pytest_asyncio.fixture
async def gmap_client():
    devices = TwoDevices()
    devices[0].add_service(gmas_service)

    await devices.setup_connection()

    assert devices.connections[0]
    assert devices.connections[1]

    devices.connections[0].encryption = 1
    devices.connections[1].encryption = 1

    peer = device.Peer(devices.connections[1])

    gmap_client = await peer.discover_service_and_create_proxy(GamingAudioServiceProxy)

    assert gmap_client
    yield gmap_client


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_init_service(gmap_client: GamingAudioServiceProxy):
    assert (
        await gmap_client.gmap_role.read_value()
        == GmapRole.UNICAST_GAME_GATEWAY
        | GmapRole.UNICAST_GAME_TERMINAL
        | GmapRole.BROADCAST_GAME_RECEIVER
        | GmapRole.BROADCAST_GAME_SENDER
    )
    assert gmap_client.ugg_features is not None
    assert await gmap_client.ugg_features.read_value() == UggFeatures.UGG_MULTISINK
    assert gmap_client.ugt_features is not None
    assert await gmap_client.ugt_features.read_value() == UgtFeatures.UGT_SOURCE
    assert gmap_client.bgr_features is not None
    assert await gmap_client.bgr_features.read_value() == BgrFeatures.BGR_MULTISINK
    assert gmap_client.bgs_features is not None
    assert await gmap_client.bgs_features.read_value() == BgsFeatures.BGS_96_KBPS



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import sys

import bumble.logging
from bumble.core import UUID
from bumble.device import Device
from bumble.rfcomm import Server, make_service_sdp_records
from bumble.transport import open_transport
from bumble.utils import AsyncRunner


# -----------------------------------------------------------------------------
def sdp_records(channel, uuid):
    service_record_handle = 0x00010001
    return {
        service_record_handle: make_service_sdp_records(
            service_record_handle, channel, UUID(uuid)
        )
    }


# -----------------------------------------------------------------------------
def on_rfcomm_session(rfcomm_session, tcp_server):
    print('*** RFComm session connected', rfcomm_session)
    tcp_server.attach_session(rfcomm_session)


# -----------------------------------------------------------------------------
class TcpServerProtocol(asyncio.Protocol):
    def __init__(self, server):
        self.server = server

    def connection_made(self, transport):
        peer_name = transport.get_extra_info('peer_name')
        print(f'<<< TCP Server: connection from {peer_name}')
        if self.server:
            self.server.tcp_transport = transport
        else:
            transport.close()

    def connection_lost(self, exc):
        print('<<< TCP Server: connection lost')
        if self.server:
            self.server.tcp_transport = None

    def data_received(self, data):
        print(f'<<< TCP Server: data received: {len(data)} bytes - {data.hex()}')
        if self.server:
            self.server.tcp_data_received(data)


# -----------------------------------------------------------------------------
class TcpServer:
    def __init__(self, port):
        self.rfcomm_session = None
        self.tcp_transport = None
        AsyncRunner.spawn(self.run(port))

    def attach_session(self, rfcomm_session):
        if self.rfcomm_session:
            self.rfcomm_session.sink = None

        self.rfcomm_session = rfcomm_session
        rfcomm_session.sink = self.rfcomm_data_received

    def rfcomm_data_received(self, data):
        print(f'<<< RFCOMM Data: {data.hex()}')
        if self.tcp_transport:
            self.tcp_transport.write(data)
        else:
            print('!!! no TCP connection, dropping data')

    def tcp_data_received(self, data):
        if self.rfcomm_session:
            self.rfcomm_session.write(data)
        else:
            print('!!! no RFComm session, dropping data')

    async def run(self, port):
        print(f'$$$ Starting TCP server on port {port}')

        server = await asyncio.get_running_loop().create_server(
            lambda: TcpServerProtocol(self), '127.0.0.1', port
        )

        async with server:
            await server.serve_forever()


# -----------------------------------------------------------------------------
async def main() -> None:
    if len(sys.argv) < 4:
        print(
            'Usage: run_rfcomm_server.py <device-config> <transport-spec> '
            '<tcp-port> [<uuid>]'
        )
        print('example: run_rfcomm_server.py classic2.json usb:0 8888')
        return

    tcp_port = int(sys.argv[3])

    if len(sys.argv) >= 5:
        uuid = sys.argv[4]
    else:
        uuid = 'E6D55659-C8B4-4B85-96BB-B1143AF6D3AE'

    print('<<< connecting to HCI...')
    async with await open_transport(sys.argv[2]) as hci_transport:
        print('<<< connected')

        # Create a device
        device = Device.from_config_file_with_hci(
            sys.argv[1], hci_transport.source, hci_transport.sink
        )
        device.classic_enabled = True

        # Create a TCP server
        tcp_server = TcpServer(tcp_port)

        # Create and register an RFComm server
        rfcomm_server = Server(device)

        # Listen for incoming DLC connections
        channel_number = rfcomm_server.listen(
            lambda session: on_rfcomm_session(session, tcp_server)
        )
        print(f'### Listening for RFComm connections on channel {channel_number}')

        # Setup the SDP to advertise this channel
        device.sdp_service_records = sdp_records(channel_number, uuid)

        # Start the controller
        await device.power_on()

        # Start being discoverable and connectable
        await device.set_discoverable(True)
        await device.set_connectable(True)

        await hci_transport.source.terminated


# -----------------------------------------------------------------------------
bumble.logging.setup_basic_logging('DEBUG')
asyncio.run(main())



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import hashlib
import os

from bumble.decoder import G722Decoder


# -----------------------------------------------------------------------------
def test_decode_file():
    decoder = G722Decoder()
    output_bytes = bytearray()

    with open(
        os.path.join(os.path.dirname(__file__), 'g722_sample.g722'), 'rb'
    ) as file:
        file_content = file.read()
        frame_length = 80
        data_length = int(len(file_content) / frame_length)

        for i in range(0, data_length):
            decoded_data = decoder.decode_frame(
                file_content[i * frame_length : i * frame_length + frame_length]
            )
            output_bytes.extend(decoded_data)

        result = hashlib.md5(output_bytes).hexdigest()
        assert result == 'b58e0cdd012d12f5633fc796c3b0fbd4'


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    test_decode_file()



from bumble.transport.grpc_protobuf.netsim import hci_packet_pb2 as _hci_packet_pb2
from bumble.transport.grpc_protobuf.netsim import startup_pb2 as _startup_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class PacketRequest(_message.Message):
    __slots__ = ("initial_info", "hci_packet", "packet")
    INITIAL_INFO_FIELD_NUMBER: _ClassVar[int]
    HCI_PACKET_FIELD_NUMBER: _ClassVar[int]
    PACKET_FIELD_NUMBER: _ClassVar[int]
    initial_info: _startup_pb2.ChipInfo
    hci_packet: _hci_packet_pb2.HCIPacket
    packet: bytes
    def __init__(self, initial_info: _Optional[_Union[_startup_pb2.ChipInfo, _Mapping]] = ..., hci_packet: _Optional[_Union[_hci_packet_pb2.HCIPacket, _Mapping]] = ..., packet: _Optional[bytes] = ...) -> None: ...

class PacketResponse(_message.Message):
    __slots__ = ("error", "hci_packet", "packet")
    ERROR_FIELD_NUMBER: _ClassVar[int]
    HCI_PACKET_FIELD_NUMBER: _ClassVar[int]
    PACKET_FIELD_NUMBER: _ClassVar[int]
    error: str
    hci_packet: _hci_packet_pb2.HCIPacket
    packet: bytes
    def __init__(self, error: _Optional[str] = ..., hci_packet: _Optional[_Union[_hci_packet_pb2.HCIPacket, _Mapping]] = ..., packet: _Optional[bytes] = ...) -> None: ...



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import json
import logging
import os
import pathlib
import tempfile
from unittest import mock

import pytest

from bumble.keys import JsonKeyStore, PairingKeys

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------

JSON1 = """
        {
            "my_namespace": {
                "14:7D:DA:4E:53:A8/P": {
                    "address_type": 0,
                    "irk": {
                        "authenticated": false,
                        "value": "e7b2543b206e4e46b44f9e51dad22bd1"
                    },
                    "link_key": {
                        "authenticated": false,
                        "value": "0745dd9691e693d9dca740f7d8dfea75"
                    },
                    "ltk": {
                        "authenticated": false,
                        "value": "d1897ee10016eb1a08e4e037fd54c683"
                    }
                }
            }
        }
        """

JSON2 = """
        {
            "my_namespace1": {
            },
            "my_namespace2": {
            }
        }
        """

JSON3 = """
        {
            "my_namespace1": {
            },
            "__DEFAULT__": {
                "14:7D:DA:4E:53:A8/P": {
                    "address_type": 0,
                    "irk": {
                        "authenticated": false,
                        "value": "e7b2543b206e4e46b44f9e51dad22bd1"
                    }
                }
            }
        }
        """


# -----------------------------------------------------------------------------
@pytest.fixture
def temporary_file():
    file = tempfile.NamedTemporaryFile(delete=False)
    file.close()
    yield file.name
    pathlib.Path(file.name).unlink()


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_basic(temporary_file):
    with open(temporary_file, mode='w', encoding='utf-8') as file:
        file.write("{}")
        file.flush()

    keystore = JsonKeyStore('my_namespace', temporary_file)

    keys = await keystore.get_all()
    assert len(keys) == 0

    keys = PairingKeys()
    await keystore.update('foo', keys)
    foo = await keystore.get('foo')
    assert foo is not None
    assert foo.ltk is None
    ltk = bytes([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15])
    keys.ltk = PairingKeys.Key(ltk)
    await keystore.update('foo', keys)
    foo = await keystore.get('foo')
    assert foo is not None
    assert foo.ltk is not None
    assert foo.ltk.value == ltk

    with open(file.name, encoding="utf-8") as json_file:
        json_data = json.load(json_file)
        assert 'my_namespace' in json_data
        assert 'foo' in json_data['my_namespace']
        assert 'ltk' in json_data['my_namespace']['foo']


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_parsing(temporary_file):
    with open(temporary_file, mode='w', encoding='utf-8') as file:
        file.write(JSON1)
        file.flush()

    keystore = JsonKeyStore('my_namespace', file.name)
    foo = await keystore.get('14:7D:DA:4E:53:A8/P')
    assert foo is not None
    assert foo.ltk.value == bytes.fromhex('d1897ee10016eb1a08e4e037fd54c683')


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_default_namespace(temporary_file):
    with open(temporary_file, mode='w', encoding='utf-8') as file:
        file.write(JSON1)
        file.flush()

    keystore = JsonKeyStore(None, file.name)
    all_keys = await keystore.get_all()
    assert len(all_keys) == 1
    name, keys = all_keys[0]
    assert name == '14:7D:DA:4E:53:A8/P'
    assert keys.irk.value == bytes.fromhex('e7b2543b206e4e46b44f9e51dad22bd1')

    with open(temporary_file, mode='w', encoding='utf-8') as file:
        file.write(JSON2)
        file.flush()

    keystore = JsonKeyStore(None, file.name)
    keys = PairingKeys()
    ltk = bytes([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15])
    keys.ltk = PairingKeys.Key(ltk)
    await keystore.update('foo', keys)
    with open(file.name, encoding="utf-8") as json_file:
        json_data = json.load(json_file)
        assert '__DEFAULT__' in json_data
        assert 'foo' in json_data['__DEFAULT__']
        assert 'ltk' in json_data['__DEFAULT__']['foo']

    with open(temporary_file, mode='w', encoding='utf-8') as file:
        file.write(JSON3)
        file.flush()

    keystore = JsonKeyStore(None, file.name)
    all_keys = await keystore.get_all()
    assert len(all_keys) == 1
    name, keys = all_keys[0]
    assert name == '14:7D:DA:4E:53:A8/P'
    assert keys.irk.value == bytes.fromhex('e7b2543b206e4e46b44f9e51dad22bd1')


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_no_filename(tmp_path):
    import platformdirs

    with mock.patch.object(platformdirs, 'user_data_path', return_value=tmp_path):
        # Case 1: no namespace, no filename
        keystore = JsonKeyStore(None, None)
        expected_directory = tmp_path / 'Pairing'
        expected_filename = expected_directory / 'keys.json'
        assert keystore.directory_name == expected_directory
        assert keystore.filename == expected_filename

        # Save some data
        keys = PairingKeys()
        ltk = bytes([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15])
        keys.ltk = PairingKeys.Key(ltk)
        await keystore.update('foo', keys)
        assert expected_filename.exists()

        # Load back
        keystore2 = JsonKeyStore(None, None)
        foo = await keystore2.get('foo')
        assert foo is not None
        assert foo.ltk.value == ltk

        # Case 2: namespace, no filename
        keystore3 = JsonKeyStore('my:namespace', None)
        # safe_name = 'my-namespace' (lower is already 'my:namespace', then replace ':' with '-')
        expected_filename3 = expected_directory / 'my-namespace.json'
        assert keystore3.filename == expected_filename3

        # Save some data
        await keystore3.update('bar', keys)
        assert expected_filename3.exists()

        # Load back
        keystore4 = JsonKeyStore('my:namespace', None)
        bar = await keystore4.get('bar')
        assert bar is not None
        assert bar.ltk.value == ltk


# -----------------------------------------------------------------------------
async def run_tests():
    await test_basic()
    await test_parsing()
    await test_default_namespace()
    await test_no_filename()


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    logging.basicConfig(level=os.environ.get('BUMBLE_LOGLEVEL', 'INFO').upper())
    asyncio.run(run_tests())



# Copyright 2021-2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import logging

import click

import bumble.logging
from bumble import transport
from bumble.drivers import rtk
from bumble.host import Host

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
def do_parse(firmware_path):
    with open(firmware_path, 'rb') as firmware_file:
        firmware_data = firmware_file.read()
        firmware = rtk.Firmware(firmware_data)
        print(
            f"Firmware: version=0x{firmware.version:08X} "
            f"project_id=0x{firmware.project_id:04X}"
        )
        for patch in firmware.patches:
            print(
                f"  Patch: chip_id=0x{patch[0]:04X}, "
                f"{len(patch[1])} bytes, "
                f"SVN Version={patch[2]:08X}"
            )


# -----------------------------------------------------------------------------
async def do_load(usb_transport, force):
    async with await transport.open_transport(usb_transport) as (
        hci_source,
        hci_sink,
    ):
        # Create a host to communicate with the device
        host = Host(hci_source, hci_sink)
        await host.reset(driver_factory=None)

        # Get the driver.
        driver = await rtk.Driver.for_host(host, force)
        if driver is None:
            # Try to see if there's already a FW image loaded
            firmware_version = await rtk.Driver.get_loaded_firmware_version(host)
            if firmware_version is None:
                print("Device not supported")
                return

            print(f"Firmware already loaded: 0x{firmware_version:08X}")
            return

        firmware_version = await driver.download_firmware()

        if firmware_version is None:
            print("Failed to load firmware")
            return

        print(f"Loaded firmware version 0x{firmware_version:08X}")


# -----------------------------------------------------------------------------
async def do_drop(usb_transport):
    async with await transport.open_transport(usb_transport) as (
        hci_source,
        hci_sink,
    ):
        # Create a host to communicate with the device
        host = Host(hci_source, hci_sink)
        await host.reset(driver_factory=None)

        # Tell the device to reset/drop any loaded patch
        await rtk.Driver.drop_firmware(host)


# -----------------------------------------------------------------------------
async def do_info(usb_transport, force):
    async with await transport.open_transport(usb_transport) as (
        hci_source,
        hci_sink,
    ):
        # Create a host to communicate with the device
        host = Host(hci_source, hci_sink)
        await host.reset(driver_factory=None)

        # Check if this is a supported device.
        if not force and not rtk.Driver.check(host):
            print("USB device not supported by this RTK driver")
            return

        # Get the driver info.
        driver_info = await rtk.Driver.driver_info_for_host(host)
        if driver_info:
            print(
                "Driver:\n"
                f"  ROM:      {driver_info.rom:04X}\n"
                f"  Firmware: {driver_info.fw_name}\n"
                f"  Config:   {driver_info.config_name}\n"
            )
        else:
            # Try to see if there's already a FW image loaded
            firmware_version = await rtk.Driver.get_loaded_firmware_version(host)
            if firmware_version is None:
                print("Device not supported")
                return

            print(f"Firmware loaded: 0x{firmware_version:08X}")


# -----------------------------------------------------------------------------
@click.group()
def main():
    bumble.logging.setup_basic_logging()


@main.command
@click.argument("firmware_path")
def parse(firmware_path):
    """Parse a firmware image."""
    do_parse(firmware_path)


@main.command
@click.argument("usb_transport")
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Load even if the USB info doesn't match",
)
def load(usb_transport, force):
    """Load a firmware image into the USB dongle."""
    asyncio.run(do_load(usb_transport, force))


@main.command
@click.argument("usb_transport")
def drop(usb_transport):
    """Drop a firmware image from the USB dongle."""
    asyncio.run(do_drop(usb_transport))


@main.command
@click.argument("usb_transport")
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Try to get the device info even if the USB info doesn't match",
)
def info(usb_transport, force):
    """Get the firmware info from a USB dongle."""
    asyncio.run(do_info(usb_transport, force))


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    main()



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import sys
from collections.abc import Callable

import bumble.logging
from bumble.core import BaseError
from bumble.device import Connection, Device
from bumble.hci import Address, LeFeatureMask
from bumble.transport import open_transport

# -----------------------------------------------------------------------------
DEFAULT_CENTRAL_ADDRESS = Address("F0:F0:F0:F0:F0:F0")
DEFAULT_PERIPHERAL_ADDRESS = Address("F1:F1:F1:F1:F1:F1")


# -----------------------------------------------------------------------------
async def run_as_central(
    device: Device,
    scenario: Callable | None,
) -> None:
    # Connect to the peripheral
    print(f'=== Connecting to {DEFAULT_PERIPHERAL_ADDRESS}...')
    connection = await device.connect(DEFAULT_PERIPHERAL_ADDRESS)
    print("=== Connected")

    if scenario is not None:
        await asyncio.sleep(1)
        await scenario(connection)

    await asyncio.get_running_loop().create_future()


async def run_as_peripheral(device: Device, scenario: Callable | None) -> None:
    # Wait for a connection from the central
    print(f'=== Advertising as {DEFAULT_PERIPHERAL_ADDRESS}...')
    await device.start_advertising(auto_restart=True)

    async def on_connection(connection: Connection) -> None:
        assert scenario is not None
        await asyncio.sleep(1)
        await scenario(connection)

    if scenario is not None:
        device.on(Device.EVENT_CONNECTION, on_connection)

    await asyncio.get_running_loop().create_future()


async def change_parameters(
    connection: Connection,
    parameter_request_procedure_supported: bool,
    subrating_supported: bool,
    shorter_connection_intervals_supported: bool,
) -> None:
    if parameter_request_procedure_supported:
        try:
            print(">>> update_parameters(7.5, 200, 0, 4000)")
            await connection.update_parameters(7.5, 200, 0, 4000)
            await asyncio.sleep(3)
        except BaseError as error:
            print(f"Error: {error}")

    if subrating_supported:
        try:
            print(">>> update_subrate(1, 2, 2, 1, 4000)")
            await connection.update_subrate(1, 2, 2, 1, 4000)
            await asyncio.sleep(3)
        except BaseError as error:
            print(f"Error: {error}")

    if shorter_connection_intervals_supported:
        try:
            print(
                ">>> update_parameters_with_subrate(7.5, 200, 1, 1, 0, 0, 4000, 5, 1000)"
            )
            await connection.update_parameters_with_subrate(
                7.5, 200, 1, 1, 0, 0, 4000, 5, 1000
            )
            await asyncio.sleep(3)
        except BaseError as error:
            print(f"Error: {error}")

        try:
            print(
                ">>> update_parameters_with_subrate(0.750, 5, 1, 1, 0, 0, 4000, 0.125, 1000)"
            )
            await connection.update_parameters_with_subrate(
                0.750, 5, 1, 1, 0, 0, 4000, 0.125, 1000
            )
            await asyncio.sleep(3)
        except BaseError as error:
            print(f"Error: {error}")

    print(">>> done")


def on_connection(connection: Connection) -> None:
    print(f"+++ Connection established: {connection}")

    def on_le_remote_features_change() -> None:
        print(f'... LE Remote Features change: {connection.peer_le_features.name}')

    connection.on(
        connection.EVENT_LE_REMOTE_FEATURES_CHANGE, on_le_remote_features_change
    )

    def on_connection_parameters_change() -> None:
        print(f'... LE Connection Parameters change: {connection.parameters}')

    connection.on(
        connection.EVENT_CONNECTION_PARAMETERS_UPDATE, on_connection_parameters_change
    )


async def main() -> None:
    if len(sys.argv) < 3:
        print(
            'Usage: run_connection_updates.py <transport-spec> '
            'central|peripheral initiator|responder'
        )
        return

    print('<<< connecting to HCI...')
    async with await open_transport(sys.argv[1]) as hci_transport:
        print('<<< connected')

        role = sys.argv[2]
        direction = sys.argv[3]
        device = Device.with_hci(
            role,
            (
                DEFAULT_CENTRAL_ADDRESS
                if role == "central"
                else DEFAULT_PERIPHERAL_ADDRESS
            ),
            hci_transport.source,
            hci_transport.sink,
        )
        device.le_subrate_enabled = True
        device.le_shorter_connection_intervals_enabled = True
        await device.power_on()

        parameter_request_procedure_supported = device.supports_le_features(
            LeFeatureMask.CONNECTION_PARAMETERS_REQUEST_PROCEDURE
        )
        print(
            "Parameters Request Procedure supported: "
            f"{parameter_request_procedure_supported}"
        )

        subrating_supported = device.supports_le_features(
            LeFeatureMask.CONNECTION_SUBRATING
        )
        print(f"Subrating supported: {subrating_supported}")

        shorter_connection_intervals_supported = device.supports_le_features(
            LeFeatureMask.SHORTER_CONNECTION_INTERVALS
        )
        print(
            "Shorter Connection Intervals supported: "
            f"{shorter_connection_intervals_supported}"
        )

        device.on(Device.EVENT_CONNECTION, on_connection)

        async def run(connection: Connection) -> None:
            await change_parameters(
                connection,
                parameter_request_procedure_supported,
                subrating_supported,
                shorter_connection_intervals_supported,
            )

        scenario = run if direction == "initiator" else None

        if role == "central":
            await run_as_central(device, scenario)
        else:
            await run_as_peripheral(device, scenario)


# -----------------------------------------------------------------------------
bumble.logging.setup_basic_logging('DEBUG')
asyncio.run(main())



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import struct

from bumble.gatt import (
    GATT_DEVICE_INFORMATION_SERVICE,
    GATT_FIRMWARE_REVISION_STRING_CHARACTERISTIC,
    GATT_HARDWARE_REVISION_STRING_CHARACTERISTIC,
    GATT_MANUFACTURER_NAME_STRING_CHARACTERISTIC,
    GATT_MODEL_NUMBER_STRING_CHARACTERISTIC,
    GATT_REGULATORY_CERTIFICATION_DATA_LIST_CHARACTERISTIC,
    GATT_SERIAL_NUMBER_STRING_CHARACTERISTIC,
    GATT_SOFTWARE_REVISION_STRING_CHARACTERISTIC,
    GATT_SYSTEM_ID_CHARACTERISTIC,
    Characteristic,
    TemplateService,
)
from bumble.gatt_adapters import (
    DelegatedCharacteristicProxyAdapter,
    UTF8CharacteristicProxyAdapter,
)
from bumble.gatt_client import CharacteristicProxy, ProfileServiceProxy, ServiceProxy


# -----------------------------------------------------------------------------
class DeviceInformationService(TemplateService):
    UUID = GATT_DEVICE_INFORMATION_SERVICE

    @staticmethod
    def pack_system_id(oui, manufacturer_id):
        return struct.pack('<Q', oui << 40 | manufacturer_id)

    @staticmethod
    def unpack_system_id(buffer):
        system_id = struct.unpack('<Q', buffer)[0]
        return (system_id >> 40, system_id & 0xFFFFFFFFFF)

    def __init__(
        self,
        manufacturer_name: str | None = None,
        model_number: str | None = None,
        serial_number: str | None = None,
        hardware_revision: str | None = None,
        firmware_revision: str | None = None,
        software_revision: str | None = None,
        system_id: tuple[int, int] | None = None,  # (OUI, Manufacturer ID)
        ieee_regulatory_certification_data_list: bytes | None = None,
        # TODO: pnp_id
    ):
        characteristics: list[Characteristic[bytes]] = [
            Characteristic(
                uuid,
                Characteristic.Properties.READ,
                Characteristic.READABLE,
                bytes(field, 'utf-8'),
            )
            for (field, uuid) in (
                (manufacturer_name, GATT_MANUFACTURER_NAME_STRING_CHARACTERISTIC),
                (model_number, GATT_MODEL_NUMBER_STRING_CHARACTERISTIC),
                (serial_number, GATT_SERIAL_NUMBER_STRING_CHARACTERISTIC),
                (hardware_revision, GATT_HARDWARE_REVISION_STRING_CHARACTERISTIC),
                (firmware_revision, GATT_FIRMWARE_REVISION_STRING_CHARACTERISTIC),
                (software_revision, GATT_SOFTWARE_REVISION_STRING_CHARACTERISTIC),
            )
            if field is not None
        ]

        if system_id is not None:
            characteristics.append(
                Characteristic(
                    GATT_SYSTEM_ID_CHARACTERISTIC,
                    Characteristic.Properties.READ,
                    Characteristic.READABLE,
                    self.pack_system_id(*system_id),
                )
            )

        if ieee_regulatory_certification_data_list is not None:
            characteristics.append(
                Characteristic(
                    GATT_REGULATORY_CERTIFICATION_DATA_LIST_CHARACTERISTIC,
                    Characteristic.Properties.READ,
                    Characteristic.READABLE,
                    ieee_regulatory_certification_data_list,
                )
            )

        super().__init__(characteristics)


# -----------------------------------------------------------------------------
class DeviceInformationServiceProxy(ProfileServiceProxy):
    SERVICE_CLASS = DeviceInformationService

    manufacturer_name: CharacteristicProxy[str] | None
    model_number: CharacteristicProxy[str] | None
    serial_number: CharacteristicProxy[str] | None
    hardware_revision: CharacteristicProxy[str] | None
    firmware_revision: CharacteristicProxy[str] | None
    software_revision: CharacteristicProxy[str] | None
    system_id: CharacteristicProxy[tuple[int, int]] | None
    ieee_regulatory_certification_data_list: CharacteristicProxy[bytes] | None

    def __init__(self, service_proxy: ServiceProxy):
        self.service_proxy = service_proxy

        for field, uuid in (
            ('manufacturer_name', GATT_MANUFACTURER_NAME_STRING_CHARACTERISTIC),
            ('model_number', GATT_MODEL_NUMBER_STRING_CHARACTERISTIC),
            ('serial_number', GATT_SERIAL_NUMBER_STRING_CHARACTERISTIC),
            ('hardware_revision', GATT_HARDWARE_REVISION_STRING_CHARACTERISTIC),
            ('firmware_revision', GATT_FIRMWARE_REVISION_STRING_CHARACTERISTIC),
            ('software_revision', GATT_SOFTWARE_REVISION_STRING_CHARACTERISTIC),
        ):
            if characteristics := service_proxy.get_characteristics_by_uuid(uuid):
                characteristic = UTF8CharacteristicProxyAdapter(characteristics[0])
            else:
                characteristic = None
            self.__setattr__(field, characteristic)

        if characteristics := service_proxy.get_characteristics_by_uuid(
            GATT_SYSTEM_ID_CHARACTERISTIC
        ):
            self.system_id = DelegatedCharacteristicProxyAdapter(
                characteristics[0],
                encode=lambda v: DeviceInformationService.pack_system_id(*v),
                decode=DeviceInformationService.unpack_system_id,
            )
        else:
            self.system_id = None

        if characteristics := service_proxy.get_characteristics_by_uuid(
            GATT_REGULATORY_CERTIFICATION_DATA_LIST_CHARACTERISTIC
        ):
            self.ieee_regulatory_certification_data_list = characteristics[0]
        else:
            self.ieee_regulatory_certification_data_list = None



# Copyright 2021-2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import secrets
import sys

import bumble.logging
from bumble import data_types
from bumble.core import AdvertisingData
from bumble.device import Device
from bumble.hci import Address
from bumble.profiles.cap import CommonAudioServiceService
from bumble.profiles.csip import CoordinatedSetIdentificationService, SirkType
from bumble.transport import open_transport


# -----------------------------------------------------------------------------
async def main() -> None:
    if len(sys.argv) < 3:
        print(
            'Usage: run_csis_servers.py <config-file> '
            '<transport-spec-for-device-1> <transport-spec-for-device-2>'
        )
        print('example: run_csis_servers.py device1.json hci-socket:0 hci-socket:1')
        return

    print('<<< connecting to HCI...')
    hci_transports = await asyncio.gather(
        open_transport(sys.argv[2]), open_transport(sys.argv[3])
    )
    print('<<< connected')

    devices = [
        Device.from_config_file_with_hci(
            sys.argv[1], hci_transport.source, hci_transport.sink
        )
        for hci_transport in hci_transports
    ]

    sirk = secrets.token_bytes(16)

    for i, device in enumerate(devices):
        device.random_address = Address(secrets.token_bytes(6))
        await device.power_on()
        csis = CoordinatedSetIdentificationService(
            set_identity_resolving_key=sirk,
            set_identity_resolving_key_type=SirkType.PLAINTEXT,
            coordinated_set_size=2,
        )
        device.add_service(CommonAudioServiceService(csis))
        advertising_data = (
            bytes(
                AdvertisingData(
                    [
                        data_types.CompleteLocalName(f'Bumble LE Audio-{i}'),
                        data_types.Flags(
                            AdvertisingData.LE_GENERAL_DISCOVERABLE_MODE_FLAG
                            | AdvertisingData.BR_EDR_HOST_FLAG
                            | AdvertisingData.BR_EDR_CONTROLLER_FLAG
                        ),
                        data_types.IncompleteListOf16BitServiceUUIDs(
                            [CoordinatedSetIdentificationService.UUID]
                        ),
                    ]
                )
            )
            + csis.get_advertising_data()
        )
        await device.create_advertising_set(advertising_data=advertising_data)

    await asyncio.gather(
        *[hci_transport.source.terminated for hci_transport in hci_transports]
    )


# -----------------------------------------------------------------------------
bumble.logging.setup_basic_logging('DEBUG')
asyncio.run(main())



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import asyncio
import logging

import click

import bumble.logging
from bumble.a2dp import (
    A2DP_MPEG_2_4_AAC_CODEC_TYPE,
    A2DP_NON_A2DP_CODEC_TYPE,
    A2DP_SBC_CODEC_TYPE,
    AacFrame,
    AacMediaCodecInformation,
    AacPacketSource,
    AacParser,
    OpusMediaCodecInformation,
    OpusPacket,
    OpusPacketSource,
    OpusParser,
    SbcFrame,
    SbcMediaCodecInformation,
    SbcPacketSource,
    SbcParser,
    make_audio_source_service_sdp_records,
)
from bumble.avdtp import (
    AVDTP_AUDIO_MEDIA_TYPE,
    AVDTP_DELAY_REPORTING_SERVICE_CATEGORY,
    MediaCodecCapabilities,
    MediaPacketPump,
    find_avdtp_service_with_connection,
)
from bumble.avdtp import Protocol as AvdtpProtocol
from bumble.avrcp import Protocol as AvrcpProtocol
from bumble.colors import color
from bumble.core import AdvertisingData, DeviceClass, PhysicalTransport
from bumble.core import ConnectionError as BumbleConnectionError
from bumble.device import Connection, Device, DeviceConfiguration
from bumble.hci import HCI_CONNECTION_ALREADY_EXISTS_ERROR, Address, HCI_Constant
from bumble.pairing import PairingConfig
from bumble.transport import open_transport
from bumble.utils import AsyncRunner

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
def a2dp_source_sdp_records():
    service_record_handle = 0x00010001
    return {
        service_record_handle: make_audio_source_service_sdp_records(
            service_record_handle
        )
    }


# -----------------------------------------------------------------------------
async def sbc_codec_capabilities(read_function) -> MediaCodecCapabilities:
    sbc_parser = SbcParser(read_function)
    sbc_frame: SbcFrame
    async for sbc_frame in sbc_parser.frames:
        # We only need the first frame
        print(color(f"SBC format: {sbc_frame}", "cyan"))
        break

    channel_mode = [
        SbcMediaCodecInformation.ChannelMode.MONO,
        SbcMediaCodecInformation.ChannelMode.DUAL_CHANNEL,
        SbcMediaCodecInformation.ChannelMode.STEREO,
        SbcMediaCodecInformation.ChannelMode.JOINT_STEREO,
    ][sbc_frame.channel_mode]
    block_length = {
        4: SbcMediaCodecInformation.BlockLength.BL_4,
        8: SbcMediaCodecInformation.BlockLength.BL_8,
        12: SbcMediaCodecInformation.BlockLength.BL_12,
        16: SbcMediaCodecInformation.BlockLength.BL_16,
    }[sbc_frame.block_count]
    subbands = {
        4: SbcMediaCodecInformation.Subbands.S_4,
        8: SbcMediaCodecInformation.Subbands.S_8,
    }[sbc_frame.subband_count]
    allocation_method = [
        SbcMediaCodecInformation.AllocationMethod.LOUDNESS,
        SbcMediaCodecInformation.AllocationMethod.SNR,
    ][sbc_frame.allocation_method]
    return MediaCodecCapabilities(
        media_type=AVDTP_AUDIO_MEDIA_TYPE,
        media_codec_type=A2DP_SBC_CODEC_TYPE,
        media_codec_information=SbcMediaCodecInformation(
            sampling_frequency=SbcMediaCodecInformation.SamplingFrequency.from_int(
                sbc_frame.sampling_frequency
            ),
            channel_mode=channel_mode,
            block_length=block_length,
            subbands=subbands,
            allocation_method=allocation_method,
            minimum_bitpool_value=2,
            maximum_bitpool_value=40,
        ),
    )


# -----------------------------------------------------------------------------
async def aac_codec_capabilities(read_function) -> MediaCodecCapabilities:
    aac_parser = AacParser(read_function)
    aac_frame: AacFrame
    async for aac_frame in aac_parser.frames:
        # We only need the first frame
        print(color(f"AAC format: {aac_frame}", "cyan"))
        break

    sampling_frequency = AacMediaCodecInformation.SamplingFrequency.from_int(
        aac_frame.sampling_frequency
    )
    channels = (
        AacMediaCodecInformation.Channels.MONO
        if aac_frame.channel_configuration == 1
        else AacMediaCodecInformation.Channels.STEREO
    )

    return MediaCodecCapabilities(
        media_type=AVDTP_AUDIO_MEDIA_TYPE,
        media_codec_type=A2DP_MPEG_2_4_AAC_CODEC_TYPE,
        media_codec_information=AacMediaCodecInformation(
            object_type=AacMediaCodecInformation.ObjectType.MPEG_2_AAC_LC,
            sampling_frequency=sampling_frequency,
            channels=channels,
            vbr=1,
            bitrate=128000,
        ),
    )


# -----------------------------------------------------------------------------
async def opus_codec_capabilities(read_function) -> MediaCodecCapabilities:
    opus_parser = OpusParser(read_function)
    opus_packet: OpusPacket
    async for opus_packet in opus_parser.packets:
        # We only need the first packet
        print(color(f"Opus format: {opus_packet}", "cyan"))
        break

    if opus_packet.channel_mode == OpusPacket.ChannelMode.MONO:
        channel_mode = OpusMediaCodecInformation.ChannelMode.MONO
    elif opus_packet.channel_mode == OpusPacket.ChannelMode.STEREO:
        channel_mode = OpusMediaCodecInformation.ChannelMode.STEREO
    else:
        channel_mode = OpusMediaCodecInformation.ChannelMode.DUAL_MONO

    if opus_packet.duration == 10:
        frame_size = OpusMediaCodecInformation.FrameSize.FS_10MS
    else:
        frame_size = OpusMediaCodecInformation.FrameSize.FS_20MS

    return MediaCodecCapabilities(
        media_type=AVDTP_AUDIO_MEDIA_TYPE,
        media_codec_type=A2DP_NON_A2DP_CODEC_TYPE,
        media_codec_information=OpusMediaCodecInformation(
            channel_mode=channel_mode,
            sampling_frequency=OpusMediaCodecInformation.SamplingFrequency.SF_48000,
            frame_size=frame_size,
        ),
    )


# -----------------------------------------------------------------------------
class Player:
    def __init__(
        self,
        transport: str,
        device_config: str | None,
        authenticate: bool,
        encrypt: bool,
    ) -> None:
        self.transport = transport
        self.device_config = device_config
        self.authenticate = authenticate
        self.encrypt = encrypt
        self.avrcp_protocol: AvrcpProtocol | None = None
        self.done: asyncio.Event | None

    async def run(self, workload) -> None:
        self.done = asyncio.Event()
        try:
            await self._run(workload)
        except Exception as error:
            print(color(f"!!! ERROR: {error}", "red"))

    async def _run(self, workload) -> None:
        async with await open_transport(self.transport) as (hci_source, hci_sink):
            # Create a device
            device_config = DeviceConfiguration()
            if self.device_config:
                device_config.load_from_file(self.device_config)
            else:
                device_config.name = "Bumble Player"
                device_config.class_of_device = DeviceClass.pack_class_of_device(
                    DeviceClass.AUDIO_SERVICE_CLASS,
                    DeviceClass.AUDIO_VIDEO_MAJOR_DEVICE_CLASS,
                    DeviceClass.AUDIO_VIDEO_UNCATEGORIZED_MINOR_DEVICE_CLASS,
                )
                device_config.keystore = "JsonKeyStore"

            device_config.classic_enabled = True
            device_config.le_enabled = False
            device_config.le_simultaneous_enabled = False
            device_config.classic_sc_enabled = False
            device_config.classic_smp_enabled = False
            device = Device.from_config_with_hci(device_config, hci_source, hci_sink)

            # Setup the SDP records to expose the SRC service
            device.sdp_service_records = a2dp_source_sdp_records()

            # Setup AVRCP
            self.avrcp_protocol = AvrcpProtocol()
            self.avrcp_protocol.listen(device)

            # Don't require MITM when pairing.
            device.pairing_config_factory = lambda connection: PairingConfig(mitm=False)

            # Start the controller
            await device.power_on()

            # Print some of the config/properties
            print(
                "Player Bluetooth Address:",
                color(
                    device.public_address.to_string(with_type_qualifier=False),
                    "yellow",
                ),
            )

            # Listen for connections
            device.on("connection", self.on_bluetooth_connection)

            # Run the workload
            try:
                await workload(device)
            except BumbleConnectionError as error:
                if error.error_code == HCI_CONNECTION_ALREADY_EXISTS_ERROR:
                    print(color("Connection already established", "blue"))
                else:
                    print(color(f"Failed to connect: {error}", "red"))

            # Wait until it is time to exit
            assert self.done is not None
            await asyncio.wait(
                [hci_source.terminated, asyncio.ensure_future(self.done.wait())],
                return_when=asyncio.FIRST_COMPLETED,
            )

    def on_bluetooth_connection(self, connection: Connection) -> None:
        print(color(f"--- Connected: {connection}", "cyan"))
        connection.on("disconnection", self.on_bluetooth_disconnection)

    def on_bluetooth_disconnection(self, reason) -> None:
        print(color(f"--- Disconnected: {HCI_Constant.error_name(reason)}", "cyan"))
        self.set_done()

    async def connect(self, device: Device, address: str) -> Connection:
        print(color(f"Connecting to {address}...", "green"))
        connection = await device.connect(address, transport=PhysicalTransport.BR_EDR)

        # Request authentication
        if self.authenticate:
            print(color("*** Authenticating...", "blue"))
            await connection.authenticate()
            print(color("*** Authenticated", "blue"))

        # Enable encryption
        if self.encrypt:
            print(color("*** Enabling encryption...", "blue"))
            await connection.encrypt()
            print(color("*** Encryption on", "blue"))

        return connection

    async def create_avdtp_protocol(self, connection: Connection) -> AvdtpProtocol:
        # Look for an A2DP service
        avdtp_version = await find_avdtp_service_with_connection(connection)
        if not avdtp_version:
            raise RuntimeError("no A2DP service found")

        print(color(f"AVDTP Version: {avdtp_version}"))

        # Create a client to interact with the remote device
        return await AvdtpProtocol.connect(connection, avdtp_version)

    async def stream_packets(
        self,
        protocol: AvdtpProtocol,
        codec_type: int,
        vendor_id: int,
        codec_id: int,
        packet_source: SbcPacketSource | AacPacketSource | OpusPacketSource,
        codec_capabilities: MediaCodecCapabilities,
    ):
        # Discover all endpoints on the remote device
        endpoints = await protocol.discover_remote_endpoints()
        for endpoint in endpoints:
            print('@@@', endpoint)

        # Select a sink
        sink = protocol.find_remote_sink_by_codec(
            AVDTP_AUDIO_MEDIA_TYPE, codec_type, vendor_id, codec_id
        )
        if sink is None:
            print(color('!!! no compatible sink found', 'red'))
            return
        print(f'### Selected sink: {sink.seid}')

        # Check if the sink supports delay reporting
        delay_reporting = False
        for capability in sink.capabilities:
            if capability.service_category == AVDTP_DELAY_REPORTING_SERVICE_CATEGORY:
                delay_reporting = True
                break

        def on_delay_report(delay: int):
            print(color(f"*** DELAY REPORT: {delay}", "blue"))

        # Adjust the codec capabilities for certain codecs
        for capability in sink.capabilities:
            if isinstance(capability, MediaCodecCapabilities):
                if isinstance(
                    codec_capabilities.media_codec_information, SbcMediaCodecInformation
                ) and isinstance(
                    capability.media_codec_information, SbcMediaCodecInformation
                ):
                    codec_capabilities.media_codec_information.minimum_bitpool_value = (
                        capability.media_codec_information.minimum_bitpool_value
                    )
                    codec_capabilities.media_codec_information.maximum_bitpool_value = (
                        capability.media_codec_information.maximum_bitpool_value
                    )
                    print(color("Source media codec:", "green"), codec_capabilities)

        # Stream the packets
        packet_pump = MediaPacketPump(packet_source.packets)
        source = protocol.add_source(codec_capabilities, packet_pump, delay_reporting)
        source.on("delay_report", on_delay_report)
        stream = await protocol.create_stream(source, sink)
        await stream.start()

        await packet_pump.wait_for_completion()

    async def discover(self, device: Device) -> None:
        @device.listens_to("inquiry_result")
        def on_inquiry_result(
            address: Address, class_of_device: int, data: AdvertisingData, rssi: int
        ) -> None:
            (
                service_classes,
                major_device_class,
                minor_device_class,
            ) = DeviceClass.split_class_of_device(class_of_device)
            separator = "\n  "
            print(f">>> {color(address.to_string(False), 'yellow')}:")
            print(f"  Device Class (raw): {class_of_device:06X}")
            major_class_name = DeviceClass.major_device_class_name(major_device_class)
            print(f"  Device Major Class: {major_class_name}")
            minor_class_name = DeviceClass.minor_device_class_name(
                major_device_class, minor_device_class
            )
            print(f"  Device Minor Class: {minor_class_name}")
            print(
                "  Device Services: "
                f"{', '.join(DeviceClass.service_class_labels(service_classes))}"
            )
            print(f"  RSSI: {rssi}")
            if data.ad_structures:
                print(f"  {data.to_string(separator)}")

        await device.start_discovery()

    async def pair(self, device: Device, address: str) -> None:
        print(color(f"Connecting to {address}...", "green"))
        connection = await device.connect(address, transport=PhysicalTransport.BR_EDR)

        print(color("Pairing...", "magenta"))
        await connection.authenticate()
        print(color("Pairing completed", "magenta"))
        self.set_done()

    async def inquire(self, device: Device, address: str) -> None:
        connection = await self.connect(device, address)
        avdtp_protocol = await self.create_avdtp_protocol(connection)

        # Discover the remote endpoints
        endpoints = await avdtp_protocol.discover_remote_endpoints()
        print(f'@@@ Found {len(list(endpoints))} endpoints')
        for endpoint in endpoints:
            print('@@@', endpoint)

        self.set_done()

    async def play(
        self,
        device: Device,
        address: str | None,
        audio_format: str,
        audio_file: str,
    ) -> None:
        if audio_format == "auto":
            if audio_file.endswith(".sbc"):
                audio_format = "sbc"
            elif audio_file.endswith(".aac") or audio_file.endswith(".adts"):
                audio_format = "aac"
            elif audio_file.endswith(".ogg"):
                audio_format = "opus"
            else:
                raise ValueError("Unable to determine audio format from file extension")

        device.on(
            "connection",
            lambda connection: AsyncRunner.spawn(on_connection(connection)),
        )

        async def on_connection(connection: Connection):
            avdtp_protocol = await self.create_avdtp_protocol(connection)

            with open(audio_file, 'rb') as input_file:
                # NOTE: this should be using asyncio file reading, but blocking reads
                # are good enough for this command line app.
                async def read_audio_data(byte_count):
                    return input_file.read(byte_count)

                # Obtain the codec capabilities from the stream
                packet_source: SbcPacketSource | AacPacketSource | OpusPacketSource
                vendor_id = 0
                codec_id = 0
                if audio_format == "sbc":
                    codec_type = A2DP_SBC_CODEC_TYPE
                    codec_capabilities = await sbc_codec_capabilities(read_audio_data)
                    packet_source = SbcPacketSource(
                        read_audio_data,
                        avdtp_protocol.l2cap_channel.peer_mtu,
                    )
                elif audio_format == "aac":
                    codec_type = A2DP_MPEG_2_4_AAC_CODEC_TYPE
                    codec_capabilities = await aac_codec_capabilities(read_audio_data)
                    packet_source = AacPacketSource(
                        read_audio_data,
                        avdtp_protocol.l2cap_channel.peer_mtu,
                    )
                else:
                    codec_type = A2DP_NON_A2DP_CODEC_TYPE
                    vendor_id = OpusMediaCodecInformation.VENDOR_ID
                    codec_id = OpusMediaCodecInformation.CODEC_ID
                    codec_capabilities = await opus_codec_capabilities(read_audio_data)
                    packet_source = OpusPacketSource(
                        read_audio_data,
                        avdtp_protocol.l2cap_channel.peer_mtu,
                    )

                # Rewind to the start
                input_file.seek(0)

                try:
                    await self.stream_packets(
                        avdtp_protocol,
                        codec_type,
                        vendor_id,
                        codec_id,
                        packet_source,
                        codec_capabilities,
                    )
                except Exception as error:
                    print(color(f"!!! Error while streaming: {error}", "red"))

            self.set_done()

        if address:
            await self.connect(device, address)
        else:
            print(color("Waiting for an incoming connection...", "magenta"))

    def set_done(self) -> None:
        if self.done:
            self.done.set()


# -----------------------------------------------------------------------------
def create_player(context) -> Player:
    return Player(
        transport=context.obj["hci_transport"],
        device_config=context.obj["device_config"],
        authenticate=context.obj["authenticate"],
        encrypt=context.obj["encrypt"],
    )


# -----------------------------------------------------------------------------
@click.group()
@click.pass_context
@click.option("--hci-transport", metavar="TRANSPORT", required=True)
@click.option("--device-config", metavar="FILENAME", help="Device configuration file")
@click.option(
    "--authenticate",
    is_flag=True,
    help="Request authentication when connecting",
    default=False,
)
@click.option(
    "--encrypt", is_flag=True, help="Request encryption when connecting", default=True
)
def player_cli(ctx, hci_transport, device_config, authenticate, encrypt):
    ctx.ensure_object(dict)
    ctx.obj["hci_transport"] = hci_transport
    ctx.obj["device_config"] = device_config
    ctx.obj["authenticate"] = authenticate
    ctx.obj["encrypt"] = encrypt


@player_cli.command("discover")
@click.pass_context
def discover(context):
    """Discover speakers or headphones"""
    player = create_player(context)
    asyncio.run(player.run(player.discover))


@player_cli.command("inquire")
@click.pass_context
@click.argument(
    "address",
    metavar="ADDRESS",
)
def inquire(context, address):
    """Connect to a speaker or headphone and inquire about their capabilities"""
    player = create_player(context)
    asyncio.run(player.run(lambda device: player.inquire(device, address)))


@player_cli.command("pair")
@click.pass_context
@click.argument(
    "address",
    metavar="ADDRESS",
)
def pair(context, address):
    """Pair with a speaker or headphone"""
    player = create_player(context)
    asyncio.run(player.run(lambda device: player.pair(device, address)))


@player_cli.command("play")
@click.pass_context
@click.option(
    "--connect",
    "address",
    metavar="ADDRESS",
    help="Address or name to connect to",
)
@click.option(
    "-f",
    "--audio-format",
    type=click.Choice(["auto", "sbc", "aac", "opus"]),
    help="Audio file format (use 'auto' to infer the format from the file extension)",
    default="auto",
)
@click.argument("audio_file")
def play(context, address, audio_format, audio_file):
    """Play and audio file"""
    player = create_player(context)
    asyncio.run(
        player.run(
            lambda device: player.play(device, address, audio_format, audio_file)
        )
    )


# -----------------------------------------------------------------------------
def main():
    bumble.logging.setup_basic_logging("WARNING")
    player_cli()


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Credentials(_message.Message):
    __slots__ = ["pem_cert_chain", "pem_private_key", "pem_root_certs"]
    PEM_CERT_CHAIN_FIELD_NUMBER: _ClassVar[int]
    PEM_PRIVATE_KEY_FIELD_NUMBER: _ClassVar[int]
    PEM_ROOT_CERTS_FIELD_NUMBER: _ClassVar[int]
    pem_cert_chain: str
    pem_private_key: str
    pem_root_certs: str
    def __init__(self, pem_root_certs: _Optional[str] = ..., pem_private_key: _Optional[str] = ..., pem_cert_chain: _Optional[str] = ...) -> None: ...

class Endpoint(_message.Message):
    __slots__ = ["required_headers", "target", "tls_credentials"]
    REQUIRED_HEADERS_FIELD_NUMBER: _ClassVar[int]
    TARGET_FIELD_NUMBER: _ClassVar[int]
    TLS_CREDENTIALS_FIELD_NUMBER: _ClassVar[int]
    required_headers: _containers.RepeatedCompositeFieldContainer[Header]
    target: str
    tls_credentials: Credentials
    def __init__(self, target: _Optional[str] = ..., tls_credentials: _Optional[_Union[Credentials, _Mapping]] = ..., required_headers: _Optional[_Iterable[_Union[Header, _Mapping]]] = ...) -> None: ...

class Header(_message.Message):
    __slots__ = ["key", "value"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: str
    def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...



# Copyright 2021-2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import json
import sys

import websockets.asyncio.server

import bumble.logging
from bumble import data_types
from bumble.core import AdvertisingData
from bumble.device import (
    AdvertisingEventProperties,
    AdvertisingParameters,
    Connection,
    Device,
    Peer,
)
from bumble.hci import CodecID, CodingFormat, OwnAddressType
from bumble.profiles.ascs import AudioStreamControlService
from bumble.profiles.bap import (
    AudioLocation,
    CodecSpecificCapabilities,
    ContextType,
    SupportedFrameDuration,
    SupportedSamplingFrequency,
    UnicastServerAdvertisingData,
)
from bumble.profiles.mcp import (
    GenericMediaControlServiceProxy,
    MediaControlPointOpcode,
    MediaControlServiceProxy,
    MediaState,
)
from bumble.profiles.pacs import PacRecord, PublishedAudioCapabilitiesService
from bumble.transport import open_transport


# -----------------------------------------------------------------------------
async def main() -> None:
    if len(sys.argv) < 3:
        print('Usage: run_mcp_client.py <config-file><transport-spec-for-device>')
        return

    print('<<< connecting to HCI...')
    async with await open_transport(sys.argv[2]) as hci_transport:
        print('<<< connected')

        device = Device.from_config_file_with_hci(
            sys.argv[1], hci_transport.source, hci_transport.sink
        )

        await device.power_on()

        # Add "placeholder" services to enable Android LEA features.
        device.add_service(
            PublishedAudioCapabilitiesService(
                supported_source_context=ContextType.PROHIBITED,
                available_source_context=ContextType.PROHIBITED,
                supported_sink_context=ContextType.MEDIA,
                available_sink_context=ContextType.MEDIA,
                sink_audio_locations=(
                    AudioLocation.FRONT_LEFT | AudioLocation.FRONT_RIGHT
                ),
                sink_pac=[
                    PacRecord(
                        coding_format=CodingFormat(CodecID.LC3),
                        codec_specific_capabilities=CodecSpecificCapabilities(
                            supported_sampling_frequencies=(
                                SupportedSamplingFrequency.FREQ_16000
                                | SupportedSamplingFrequency.FREQ_32000
                                | SupportedSamplingFrequency.FREQ_48000
                            ),
                            supported_frame_durations=(
                                SupportedFrameDuration.DURATION_10000_US_SUPPORTED
                            ),
                            supported_audio_channel_count=[1, 2],
                            min_octets_per_codec_frame=0,
                            max_octets_per_codec_frame=320,
                            supported_max_codec_frames_per_sdu=2,
                        ),
                    ),
                ],
            )
        )
        device.add_service(AudioStreamControlService(device, sink_ase_id=[1]))

        ws: websockets.asyncio.server.ServerConnection | None = None
        mcp: MediaControlServiceProxy | None = None

        advertising_data = bytes(
            AdvertisingData(
                [
                    data_types.CompleteLocalName('Bumble LE Audio'),
                    data_types.Flags(AdvertisingData.LE_GENERAL_DISCOVERABLE_MODE_FLAG),
                    data_types.IncompleteListOf16BitServiceUUIDs(
                        [PublishedAudioCapabilitiesService.UUID]
                    ),
                ]
            )
        ) + bytes(UnicastServerAdvertisingData())

        await device.create_advertising_set(
            advertising_parameters=AdvertisingParameters(
                advertising_event_properties=AdvertisingEventProperties(),
                own_address_type=OwnAddressType.RANDOM,
                primary_advertising_interval_max=100,
                primary_advertising_interval_min=100,
            ),
            advertising_data=advertising_data,
            auto_restart=True,
        )

        def on_media_state(media_state: MediaState) -> None:
            if ws:
                asyncio.create_task(
                    ws.send(json.dumps({'media_state': media_state.name}))
                )

        def on_track_title(title: str) -> None:
            if ws:
                asyncio.create_task(ws.send(json.dumps({'title': title})))

        def on_track_duration(duration: int) -> None:
            if ws:
                asyncio.create_task(ws.send(json.dumps({'duration': duration})))

        def on_track_position(position: int) -> None:
            if ws:
                asyncio.create_task(ws.send(json.dumps({'position': position})))

        def on_connection(connection: Connection) -> None:
            async def on_connection_async():
                async with Peer(connection) as peer:
                    nonlocal mcp
                    mcp = peer.create_service_proxy(MediaControlServiceProxy)
                    if not mcp:
                        mcp = peer.create_service_proxy(GenericMediaControlServiceProxy)
                    mcp.on('media_state', on_media_state)
                    mcp.on('track_title', on_track_title)
                    mcp.on('track_duration', on_track_duration)
                    mcp.on('track_position', on_track_position)
                    await mcp.subscribe_characteristics()

            connection.cancel_on_disconnection(on_connection_async())

        device.on('connection', on_connection)

        async def serve(websocket: websockets.asyncio.server.ServerConnection):
            nonlocal ws
            ws = websocket
            async for message in websocket:
                request = json.loads(message)
                if mcp:
                    await mcp.write_control_point(
                        MediaControlPointOpcode(request['opcode'])
                    )
            ws = None

        await websockets.asyncio.server.serve(serve, 'localhost', 8989)

        await hci_transport.source.terminated


# -----------------------------------------------------------------------------
bumble.logging.setup_basic_logging('DEBUG')
asyncio.run(main())



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import io
import json
import logging
import sys
from collections.abc import Iterable

import websockets.asyncio.server

import bumble.core
import bumble.logging
from bumble import hci, hfp, rfcomm
from bumble.core import PhysicalTransport
from bumble.device import Device, ScoLink
from bumble.transport import open_transport

logger = logging.getLogger(__name__)

ws: websockets.asyncio.server.ServerConnection | None = None
ag_protocol: hfp.AgProtocol | None = None
source_file: io.BufferedReader | None = None


def _default_configuration() -> hfp.AgConfiguration:
    return hfp.AgConfiguration(
        supported_ag_features=[
            hfp.AgFeature.HF_INDICATORS,
            hfp.AgFeature.IN_BAND_RING_TONE_CAPABILITY,
            hfp.AgFeature.REJECT_CALL,
            hfp.AgFeature.CODEC_NEGOTIATION,
            hfp.AgFeature.ESCO_S4_SETTINGS_SUPPORTED,
            hfp.AgFeature.ENHANCED_CALL_STATUS,
        ],
        supported_ag_indicators=[
            hfp.AgIndicatorState.call(),
            hfp.AgIndicatorState.callsetup(),
            hfp.AgIndicatorState.callheld(),
            hfp.AgIndicatorState.service(),
            hfp.AgIndicatorState.signal(),
            hfp.AgIndicatorState.roam(),
            hfp.AgIndicatorState.battchg(),
        ],
        supported_hf_indicators=[
            hfp.HfIndicator.ENHANCED_SAFETY,
            hfp.HfIndicator.BATTERY_LEVEL,
        ],
        supported_ag_call_hold_operations=[],
        supported_audio_codecs=[hfp.AudioCodec.CVSD, hfp.AudioCodec.MSBC],
    )


def send_message(type: str, **kwargs) -> None:
    if ws:
        asyncio.create_task(ws.send(json.dumps({'type': type, **kwargs})))


def on_speaker_volume(level: int):
    send_message(type='speaker_volume', level=level)


def on_microphone_volume(level: int):
    send_message(type='microphone_volume', level=level)


def on_supported_audio_codecs(codecs: Iterable[hfp.AudioCodec]):
    send_message(type='supported_audio_codecs', codecs=[codec.name for codec in codecs])


def on_sco_state_change(codec: int):
    if codec == hfp.AudioCodec.CVSD:
        sample_rate = 8000
    elif codec == hfp.AudioCodec.MSBC:
        sample_rate = 16000
    else:
        sample_rate = 0

    send_message(type='sco_state_change', sample_rate=sample_rate)


def on_sco_packet(packet: hci.HCI_SynchronousDataPacket):
    if ws:
        asyncio.create_task(ws.send(packet.data))
    if source_file and (pcm_data := source_file.read(packet.data_total_length)):
        assert ag_protocol
        host = ag_protocol.dlc.multiplexer.l2cap_channel.connection.device.host
        host.send_sco_sdu(
            connection_handle=packet.connection_handle,
            sdu=pcm_data,
        )


def on_hfp_state_change(connected: bool):
    send_message(type='hfp_state_change', connected=connected)


async def ws_server(ws_client: websockets.asyncio.server.ServerConnection):
    global ws
    ws = ws_client

    async for message in ws_client:
        if not ag_protocol:
            continue

        json_message = json.loads(message)
        message_type = json_message['type']
        connection = ag_protocol.dlc.multiplexer.l2cap_channel.connection
        device = connection.device

        try:
            if message_type == 'at_response':
                ag_protocol.send_response(json_message['response'])
            elif message_type == 'ag_indicator':
                ag_protocol.update_ag_indicator(
                    hfp.AgIndicator(json_message['indicator']),
                    int(json_message['value']),
                )
            elif message_type == 'negotiate_codec':
                codec = hfp.AudioCodec(int(json_message['codec']))
                await ag_protocol.negotiate_codec(codec)
            elif message_type == 'connect_sco':
                if ag_protocol.active_codec == hfp.AudioCodec.CVSD:
                    esco_param = hfp.ESCO_PARAMETERS[
                        hfp.DefaultCodecParameters.ESCO_CVSD_S4
                    ]
                elif ag_protocol.active_codec == hfp.AudioCodec.MSBC:
                    esco_param = hfp.ESCO_PARAMETERS[
                        hfp.DefaultCodecParameters.ESCO_MSBC_T2
                    ]
                else:
                    raise ValueError(f'Unsupported codec {codec}')

                await device.send_command(
                    hci.HCI_Enhanced_Setup_Synchronous_Connection_Command(
                        connection_handle=connection.handle, **esco_param.asdict()
                    )
                )
            elif message_type == 'disconnect_sco':
                # Copy the values to avoid iteration error.
                for sco_link in list(device.sco_links.values()):
                    await sco_link.disconnect()
            elif message_type == 'update_calls':
                ag_protocol.calls = [
                    hfp.CallInfo(
                        index=int(call['index']),
                        direction=hfp.CallInfoDirection(int(call['direction'])),
                        status=hfp.CallInfoStatus(int(call['status'])),
                        number=call['number'],
                        multi_party=hfp.CallInfoMultiParty.NOT_IN_CONFERENCE,
                        mode=hfp.CallInfoMode.VOICE,
                    )
                    for call in json_message['calls']
                ]

        except Exception as e:
            send_message(type='error', message=e)


# -----------------------------------------------------------------------------
async def main() -> None:
    if len(sys.argv) < 3:
        print(
            'Usage: run_hfp_gateway.py <device-config> <transport-spec> '
            '[bluetooth-address] [wav-file-for-source]'
        )
        print(
            'example: run_hfp_gateway.py hfp_gateway.json usb:0 E1:CA:72:48:C4:E8 sample.wav'
        )
        return

    print('<<< connecting to HCI...')
    async with await open_transport(sys.argv[2]) as hci_transport:
        print('<<< connected')

        # Create a device
        device = Device.from_config_file_with_hci(
            sys.argv[1], hci_transport.source, hci_transport.sink
        )
        device.classic_enabled = True
        await device.power_on()

        rfcomm_server = rfcomm.Server(device)
        configuration = _default_configuration()

        def on_dlc(dlc: rfcomm.DLC):
            global ag_protocol
            ag_protocol = hfp.AgProtocol(dlc, configuration)
            ag_protocol.on('speaker_volume', on_speaker_volume)
            ag_protocol.on('microphone_volume', on_microphone_volume)
            ag_protocol.on('supported_audio_codecs', on_supported_audio_codecs)
            on_hfp_state_change(True)
            dlc.multiplexer.l2cap_channel.on(
                'close', lambda: on_hfp_state_change(False)
            )

        channel = rfcomm_server.listen(on_dlc)
        device.sdp_service_records = {
            1: hfp.make_ag_sdp_records(1, channel, configuration)
        }

        def on_sco_connection(sco_link: ScoLink):
            assert ag_protocol
            on_sco_state_change(ag_protocol.active_codec)
            sco_link.on('disconnection', lambda _: on_sco_state_change(0))
            sco_link.sink = on_sco_packet

        device.on('sco_connection', on_sco_connection)
        if len(sys.argv) >= 4:
            # Connect to a peer
            target_address = sys.argv[3]
            print(f'=== Connecting to {target_address}...')
            connection = await device.connect(
                target_address, transport=PhysicalTransport.BR_EDR
            )
            print(f'=== Connected to {connection.peer_address}!')

            # Get a list of all the Handsfree services (should only be 1)
            if not (hfp_record := await hfp.find_hf_sdp_record(connection)):
                print('!!! no service found')
                return

            # Pick the first one
            channel, version, hf_sdp_features = hfp_record
            print(f'HF version: {version}')
            print(f'HF features: {hf_sdp_features.name}')

            # Request authentication
            print('*** Authenticating...')
            await connection.authenticate()
            print('*** Authenticated')

            # Enable encryption
            print('*** Enabling encryption...')
            await connection.encrypt()
            print('*** Encryption on')

            # Create a client and start it
            print('@@@ Starting to RFCOMM client...')
            rfcomm_client = rfcomm.Client(connection)
            rfcomm_mux = await rfcomm_client.start()
            print('@@@ Started')

            print(f'### Opening session for channel {channel}...')
            try:
                session = await rfcomm_mux.open_dlc(channel)
                print('### Session open', session)
            except bumble.core.ConnectionError as error:
                print(f'### Session open failed: {error}')
                await rfcomm_mux.disconnect()
                print('@@@ Disconnected from RFCOMM server')
                return

            on_dlc(session)

        await websockets.asyncio.server.serve(ws_server, port=8888)

        if len(sys.argv) >= 5:
            global source_file
            source_file = open(sys.argv[4], 'rb')
            # Skip header
            source_file.seek(44)

        await hci_transport.source.terminated


# -----------------------------------------------------------------------------
bumble.logging.setup_basic_logging('DEBUG')
asyncio.run(main())



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio

import click

import bumble.logging
from bumble import l2cap
from bumble.colors import color
from bumble.device import Device
from bumble.hci import HCI_Constant
from bumble.transport import open_transport
from bumble.utils import FlowControlAsyncPipe


# -----------------------------------------------------------------------------
class ServerBridge:
    """
    L2CAP CoC server bridge: waits for a peer to connect an L2CAP CoC channel
    on a specified PSM. When the connection is made, the bridge connects a TCP
    socket to a remote host and bridges the data in both directions, with flow
    control.
    When the L2CAP CoC channel is closed, the bridge disconnects the TCP socket
    and waits for a new L2CAP CoC channel to be connected.
    When the TCP connection is closed by the TCP server, XXXX
    """

    def __init__(self, psm, max_credits, mtu, mps, tcp_host, tcp_port):
        self.psm = psm
        self.max_credits = max_credits
        self.mtu = mtu
        self.mps = mps
        self.tcp_host = tcp_host
        self.tcp_port = tcp_port

    async def start(self, device: Device) -> None:
        # Listen for incoming L2CAP channel connections
        device.create_l2cap_server(
            spec=l2cap.LeCreditBasedChannelSpec(
                psm=self.psm, mtu=self.mtu, mps=self.mps, max_credits=self.max_credits
            ),
            handler=self.on_channel,
        )
        print(
            color(f'### Listening for channel connection on PSM {self.psm}', 'yellow')
        )

        def on_ble_connection(connection):
            def on_ble_disconnection(reason):
                print(
                    color('@@@ Bluetooth disconnection:', 'red'),
                    HCI_Constant.error_name(reason),
                )

            print(color('@@@ Bluetooth connection:', 'green'), connection)
            connection.on('disconnection', on_ble_disconnection)

        device.on('connection', on_ble_connection)

        await device.start_advertising(auto_restart=True)

    # Called when a new L2CAP connection is established
    def on_channel(self, l2cap_channel):
        print(color('*** L2CAP channel:', 'cyan'), l2cap_channel)

        class Pipe:
            def __init__(self, bridge, l2cap_channel):
                self.bridge = bridge
                self.tcp_transport = None
                self.l2cap_channel = l2cap_channel

                l2cap_channel.on('close', self.on_l2cap_close)
                l2cap_channel.sink = self.on_channel_sdu

            async def connect_to_tcp(self):
                # Connect to the TCP server
                print(
                    color(
                        f'### Connecting to TCP {self.bridge.tcp_host}:'
                        f'{self.bridge.tcp_port}...',
                        'yellow',
                    )
                )

                class TcpClientProtocol(asyncio.Protocol):
                    def __init__(self, pipe):
                        self.pipe = pipe

                    def connection_lost(self, exc):
                        print(color(f'!!! TCP connection lost: {exc}', 'red'))
                        if self.pipe.l2cap_channel is not None:
                            asyncio.create_task(self.pipe.l2cap_channel.disconnect())

                    def data_received(self, data):
                        print(color(f'<<< [TCP DATA]: {len(data)} bytes', 'blue'))
                        self.pipe.l2cap_channel.write(data)

                try:
                    (
                        self.tcp_transport,
                        _,
                    ) = await asyncio.get_running_loop().create_connection(
                        lambda: TcpClientProtocol(self),
                        host=self.bridge.tcp_host,
                        port=self.bridge.tcp_port,
                    )
                    print(color('### Connected', 'green'))
                except Exception as error:
                    print(color(f'!!! Connection failed: {error}', 'red'))
                    await self.l2cap_channel.disconnect()

            def on_l2cap_close(self):
                print(color('*** L2CAP channel closed', 'red'))
                self.l2cap_channel = None
                if self.tcp_transport is not None:
                    self.tcp_transport.close()

            def on_channel_sdu(self, sdu):
                print(color(f'<<< [L2CAP SDU]: {len(sdu)} bytes', 'cyan'))
                if self.tcp_transport is None:
                    print(color('!!! TCP socket not open, dropping', 'red'))
                    return
                self.tcp_transport.write(sdu)

        pipe = Pipe(self, l2cap_channel)

        asyncio.create_task(pipe.connect_to_tcp())


# -----------------------------------------------------------------------------
class ClientBridge:
    """
    L2CAP CoC client bridge: connects to a BLE device, then waits for an inbound
    TCP connection on a specified port number. When a TCP client connects, an
    L2CAP CoC channel connection to the BLE device is established, and the data
    is bridged in both directions, with flow control.
    When the TCP connection is closed by the client, the L2CAP CoC channel is
    disconnected, but the connection to the BLE device remains, ready for a new
    TCP client to connect.
    When the L2CAP CoC channel is closed, XXXX
    """

    READ_CHUNK_SIZE = 4096

    def __init__(self, psm, max_credits, mtu, mps, address, tcp_host, tcp_port):
        self.psm = psm
        self.max_credits = max_credits
        self.mtu = mtu
        self.mps = mps
        self.address = address
        self.tcp_host = tcp_host
        self.tcp_port = tcp_port

    async def start(self, device):
        print(color(f'### Connecting to {self.address}...', 'yellow'))
        connection = await device.connect(self.address)
        print(color('### Connected', 'green'))

        # Called when the BLE connection is disconnected
        def on_ble_disconnection(reason):
            print(
                color('@@@ Bluetooth disconnection:', 'red'),
                HCI_Constant.error_name(reason),
            )

        connection.on('disconnection', on_ble_disconnection)

        # Called when a TCP connection is established
        async def on_tcp_connection(reader, writer):
            peer_name = writer.get_extra_info('peer_name')
            print(color(f'<<< TCP connection from {peer_name}', 'magenta'))

            def on_channel_sdu(sdu):
                print(color(f'<<< [L2CAP SDU]: {len(sdu)} bytes', 'cyan'))
                l2cap_to_tcp_pipe.write(sdu)

            def on_l2cap_close():
                print(color('*** L2CAP channel closed', 'red'))
                l2cap_to_tcp_pipe.stop()
                writer.close()

            # Connect a new L2CAP channel
            print(color(f'>>> Opening L2CAP channel on PSM = {self.psm}', 'yellow'))
            try:
                l2cap_channel = await connection.create_l2cap_channel(
                    spec=l2cap.LeCreditBasedChannelSpec(
                        psm=self.psm,
                        max_credits=self.max_credits,
                        mtu=self.mtu,
                        mps=self.mps,
                    )
                )
                print(color('*** L2CAP channel:', 'cyan'), l2cap_channel)
            except Exception as error:
                print(color(f'!!! Connection failed: {error}', 'red'))
                writer.close()
                return

            l2cap_channel.sink = on_channel_sdu
            l2cap_channel.on('close', on_l2cap_close)

            # Start a flow control pipe from L2CAP to TCP
            l2cap_to_tcp_pipe = FlowControlAsyncPipe(
                l2cap_channel.pause_reading,
                l2cap_channel.resume_reading,
                writer.write,
                writer.drain,
            )
            l2cap_to_tcp_pipe.start()

            # Pipe data from TCP to L2CAP
            while True:
                try:
                    data = await reader.read(self.READ_CHUNK_SIZE)

                    if len(data) == 0:
                        print(color('!!! End of stream', 'red'))
                        await l2cap_channel.disconnect()
                        return

                    print(color(f'<<< [TCP DATA]: {len(data)} bytes', 'blue'))
                    l2cap_channel.write(data)
                    await l2cap_channel.drain()
                except Exception as error:
                    print(f'!!! Exception: {error}')
                    break

            writer.close()
            print(color('~~~ Bye bye', 'magenta'))

        await asyncio.start_server(
            on_tcp_connection,
            host=self.tcp_host if self.tcp_host != '_' else None,
            port=self.tcp_port,
        )
        print(
            color(
                f'### Listening for TCP connections on port {self.tcp_port}', 'magenta'
            )
        )


# -----------------------------------------------------------------------------
async def run(device_config, hci_transport, bridge):
    print('<<< connecting to HCI...')
    async with await open_transport(hci_transport) as (hci_source, hci_sink):
        print('<<< connected')

        device = Device.from_config_file_with_hci(device_config, hci_source, hci_sink)

        # Let's go
        await device.power_on()
        await bridge.start(device)

        # Wait until the transport terminates
        await hci_source.terminated


# -----------------------------------------------------------------------------
@click.group()
@click.pass_context
@click.option('--device-config', help='Device configuration file', required=True)
@click.option('--hci-transport', help='HCI transport', required=True)
@click.option('--psm', help='PSM for L2CAP', type=int, default=1234)
@click.option(
    '--l2cap-max-credits',
    help='Maximum L2CAP Credits',
    type=click.IntRange(1, 65535),
    default=128,
)
@click.option(
    '--l2cap-mtu',
    help='L2CAP MTU',
    type=click.IntRange(
        l2cap.L2CAP_LE_CREDIT_BASED_CONNECTION_MIN_MTU,
        l2cap.L2CAP_LE_CREDIT_BASED_CONNECTION_MAX_MTU,
    ),
    default=1024,
)
@click.option(
    '--l2cap-mps',
    help='L2CAP MPS',
    type=click.IntRange(
        l2cap.L2CAP_LE_CREDIT_BASED_CONNECTION_MIN_MPS,
        l2cap.L2CAP_LE_CREDIT_BASED_CONNECTION_MAX_MPS,
    ),
    default=1024,
)
def cli(
    context,
    device_config,
    hci_transport,
    psm,
    l2cap_max_credits,
    l2cap_mtu,
    l2cap_mps,
):
    context.ensure_object(dict)
    context.obj['device_config'] = device_config
    context.obj['hci_transport'] = hci_transport
    context.obj['psm'] = psm
    context.obj['max_credits'] = l2cap_max_credits
    context.obj['mtu'] = l2cap_mtu
    context.obj['mps'] = l2cap_mps


# -----------------------------------------------------------------------------
@cli.command()
@click.pass_context
@click.option('--tcp-host', help='TCP host', default='localhost')
@click.option('--tcp-port', help='TCP port', default=9544)
def server(context, tcp_host, tcp_port):
    bridge = ServerBridge(
        context.obj['psm'],
        context.obj['max_credits'],
        context.obj['mtu'],
        context.obj['mps'],
        tcp_host,
        tcp_port,
    )
    asyncio.run(run(context.obj['device_config'], context.obj['hci_transport'], bridge))


# -----------------------------------------------------------------------------
@cli.command()
@click.pass_context
@click.argument('bluetooth-address')
@click.option('--tcp-host', help='TCP host', default='_')
@click.option('--tcp-port', help='TCP port', default=9543)
def client(context, bluetooth_address, tcp_host, tcp_port):
    bridge = ClientBridge(
        context.obj['psm'],
        context.obj['max_credits'],
        context.obj['mtu'],
        context.obj['mps'],
        bluetooth_address,
        tcp_host,
        tcp_port,
    )
    asyncio.run(run(context.obj['device_config'], context.obj['hci_transport'], bridge))


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    bumble.logging.setup_basic_logging('WARNING')
    cli(obj={})  # pylint: disable=no-value-for-parameter



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from collections.abc import Callable

from bumble import device, gatt, gatt_adapters, gatt_client


# -----------------------------------------------------------------------------
class BatteryService(gatt.TemplateService):
    UUID = gatt.GATT_BATTERY_SERVICE
    BATTERY_LEVEL_FORMAT = 'B'

    battery_level_characteristic: gatt.Characteristic[int]

    def __init__(self, read_battery_level: Callable[[device.Connection], int]) -> None:
        self.battery_level_characteristic = gatt_adapters.PackedCharacteristicAdapter(
            gatt.Characteristic(
                gatt.GATT_BATTERY_LEVEL_CHARACTERISTIC,
                properties=(
                    gatt.Characteristic.Properties.READ
                    | gatt.Characteristic.Properties.NOTIFY
                ),
                permissions=gatt.Characteristic.READABLE,
                value=gatt.CharacteristicValue(read=read_battery_level),
            ),
            pack_format=BatteryService.BATTERY_LEVEL_FORMAT,
        )
        super().__init__([self.battery_level_characteristic])


# -----------------------------------------------------------------------------
class BatteryServiceProxy(gatt_client.ProfileServiceProxy):
    SERVICE_CLASS = BatteryService

    battery_level: gatt_client.CharacteristicProxy[int]

    def __init__(self, service_proxy: gatt_client.ServiceProxy) -> None:
        self.service_proxy = service_proxy

        self.battery_level = gatt_adapters.PackedCharacteristicProxyAdapter(
            service_proxy.get_required_characteristic_by_uuid(
                gatt.GATT_BATTERY_LEVEL_CHARACTERISTIC
            ),
            pack_format=BatteryService.BATTERY_LEVEL_FORMAT,
        )



# Copyright 2021-2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import logging

from bumble.transport.common import StreamPacketSink, StreamPacketSource, Transport

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
async def open_unix_client_transport(spec: str) -> Transport:
    '''Open a UNIX socket client transport.

    The parameter is the path of unix socket. For abstract socket, the first character
    needs to be '@'.

    Example:
        * /tmp/hci.socket
        * @hci_socket
    '''

    class UnixPacketSource(StreamPacketSource):
        def connection_lost(self, exc):
            logger.debug(f'connection lost: {exc}')
            self.on_transport_lost()

    # For abstract socket, the first character should be null character.
    if spec.startswith('@'):
        spec = '\0' + spec[1:]

    (
        unix_transport,
        packet_source,
    ) = await asyncio.get_running_loop().create_unix_connection(UnixPacketSource, spec)
    packet_sink = StreamPacketSink(unix_transport)

    return Transport(packet_source, packet_sink)


# -----------------------------------------------------------------------------
async def open_unix_server_transport(spec: str) -> Transport:
    '''Open a UNIX socket server transport.

    The parameter is the path of unix socket. For abstract socket, the first character
    needs to be '@'.

    Example:
        * /tmp/hci.socket
        * @hci_socket
    '''
    # For abstract socket, the first character should be null character.
    if spec.startswith('@'):
        spec = '\0' + spec[1:]

    class UnixServerTransport(Transport):
        def __init__(self, source, sink, server):
            self.server = server
            super().__init__(source, sink)

        async def close(self):
            await super().close()

    class UnixServerProtocol(asyncio.BaseProtocol):
        def __init__(self, packet_source, packet_sink):
            self.packet_source = packet_source
            self.packet_sink = packet_sink

        # Called when a new connection is established
        def connection_made(self, transport):
            peer_name = transport.get_extra_info('peer_name')
            logger.debug('connection from %s', peer_name)
            self.packet_sink.transport = transport

        # Called when the client is disconnected
        def connection_lost(self, error):
            logger.debug('connection lost: %s', error)
            self.packet_sink.transport = None

        def eof_received(self):
            logger.debug('connection end')
            self.packet_sink.transport = None

        # Called when data is received on the socket
        def data_received(self, data):
            self.packet_source.data_received(data)

    class UnixServerPacketSink:
        def __init__(self):
            self.transport = None

        def on_packet(self, packet):
            if self.transport:
                self.transport.write(packet)
            else:
                logger.debug('no client, dropping packet')

    packet_source = StreamPacketSource()
    packet_sink = UnixServerPacketSink()
    server = await asyncio.get_running_loop().create_unix_server(
        lambda: UnixServerProtocol(packet_source, packet_sink), spec
    )

    return UnixServerTransport(packet_source, packet_sink, server)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for

"""LE Audio - Broadcast Audio Scan Service"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import dataclasses
import logging
import struct
from collections.abc import Sequence
from typing import ClassVar

from bumble import core, device, gatt, gatt_adapters, gatt_client, hci, utils

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Constants
# -----------------------------------------------------------------------------
class ApplicationError(utils.OpenIntEnum):
    OPCODE_NOT_SUPPORTED = 0x80
    INVALID_SOURCE_ID = 0x81


# -----------------------------------------------------------------------------
def encode_subgroups(subgroups: Sequence[SubgroupInfo]) -> bytes:
    return bytes([len(subgroups)]) + b"".join(
        struct.pack("<IB", subgroup.bis_sync, len(subgroup.metadata))
        + subgroup.metadata
        for subgroup in subgroups
    )


def decode_subgroups(data: bytes) -> list[SubgroupInfo]:
    num_subgroups = data[0]
    offset = 1
    subgroups = []
    for _ in range(num_subgroups):
        bis_sync = struct.unpack("<I", data[offset : offset + 4])[0]
        metadata_length = data[offset + 4]
        metadata = data[offset + 5 : offset + 5 + metadata_length]
        offset += 5 + metadata_length
        subgroups.append(SubgroupInfo(bis_sync, metadata))

    return subgroups


# -----------------------------------------------------------------------------
class PeriodicAdvertisingSyncParams(utils.OpenIntEnum):
    DO_NOT_SYNCHRONIZE_TO_PA = 0x00
    SYNCHRONIZE_TO_PA_PAST_AVAILABLE = 0x01
    SYNCHRONIZE_TO_PA_PAST_NOT_AVAILABLE = 0x02


@dataclasses.dataclass
class SubgroupInfo:
    ANY_BIS: ClassVar[int] = 0xFFFFFFFF

    bis_sync: int
    metadata: bytes


class ControlPointOperation:
    class OpCode(utils.OpenIntEnum):
        REMOTE_SCAN_STOPPED = 0x00
        REMOTE_SCAN_STARTED = 0x01
        ADD_SOURCE = 0x02
        MODIFY_SOURCE = 0x03
        SET_BROADCAST_CODE = 0x04
        REMOVE_SOURCE = 0x05

    op_code: OpCode
    parameters: bytes

    @classmethod
    def from_bytes(cls, data: bytes) -> ControlPointOperation:
        op_code = data[0]

        if op_code == cls.OpCode.REMOTE_SCAN_STOPPED:
            return RemoteScanStoppedOperation()

        if op_code == cls.OpCode.REMOTE_SCAN_STARTED:
            return RemoteScanStartedOperation()

        if op_code == cls.OpCode.ADD_SOURCE:
            return AddSourceOperation.from_parameters(data[1:])

        if op_code == cls.OpCode.MODIFY_SOURCE:
            return ModifySourceOperation.from_parameters(data[1:])

        if op_code == cls.OpCode.SET_BROADCAST_CODE:
            return SetBroadcastCodeOperation.from_parameters(data[1:])

        if op_code == cls.OpCode.REMOVE_SOURCE:
            return RemoveSourceOperation.from_parameters(data[1:])

        raise core.InvalidArgumentError("invalid op code")

    def __init__(self, op_code: OpCode, parameters: bytes = b"") -> None:
        self.op_code = op_code
        self.parameters = parameters

    def __bytes__(self) -> bytes:
        return bytes([self.op_code]) + self.parameters


class RemoteScanStoppedOperation(ControlPointOperation):
    def __init__(self) -> None:
        super().__init__(ControlPointOperation.OpCode.REMOTE_SCAN_STOPPED)


class RemoteScanStartedOperation(ControlPointOperation):
    def __init__(self) -> None:
        super().__init__(ControlPointOperation.OpCode.REMOTE_SCAN_STARTED)


class AddSourceOperation(ControlPointOperation):
    @classmethod
    def from_parameters(cls, parameters: bytes) -> AddSourceOperation:
        instance = cls.__new__(cls)
        instance.op_code = ControlPointOperation.OpCode.ADD_SOURCE
        instance.parameters = parameters
        instance.advertiser_address = hci.Address.parse_address_preceded_by_type(
            parameters, 1
        )[1]
        instance.advertising_sid = parameters[7]
        instance.broadcast_id = int.from_bytes(parameters[8:11], "little")
        instance.pa_sync = PeriodicAdvertisingSyncParams(parameters[11])
        instance.pa_interval = struct.unpack("<H", parameters[12:14])[0]
        instance.subgroups = decode_subgroups(parameters[14:])
        return instance

    def __init__(
        self,
        advertiser_address: hci.Address,
        advertising_sid: int,
        broadcast_id: int,
        pa_sync: PeriodicAdvertisingSyncParams,
        pa_interval: int,
        subgroups: Sequence[SubgroupInfo],
    ) -> None:
        super().__init__(
            ControlPointOperation.OpCode.ADD_SOURCE,
            struct.pack(
                "<B6sB3sBH",
                advertiser_address.address_type,
                bytes(advertiser_address),
                advertising_sid,
                broadcast_id.to_bytes(3, "little"),
                pa_sync,
                pa_interval,
            )
            + encode_subgroups(subgroups),
        )
        self.advertiser_address = advertiser_address
        self.advertising_sid = advertising_sid
        self.broadcast_id = broadcast_id
        self.pa_sync = pa_sync
        self.pa_interval = pa_interval
        self.subgroups = list(subgroups)


class ModifySourceOperation(ControlPointOperation):
    @classmethod
    def from_parameters(cls, parameters: bytes) -> ModifySourceOperation:
        instance = cls.__new__(cls)
        instance.op_code = ControlPointOperation.OpCode.MODIFY_SOURCE
        instance.parameters = parameters
        instance.source_id = parameters[0]
        instance.pa_sync = PeriodicAdvertisingSyncParams(parameters[1])
        instance.pa_interval = struct.unpack("<H", parameters[2:4])[0]
        instance.subgroups = decode_subgroups(parameters[4:])
        return instance

    def __init__(
        self,
        source_id: int,
        pa_sync: PeriodicAdvertisingSyncParams,
        pa_interval: int,
        subgroups: Sequence[SubgroupInfo],
    ) -> None:
        super().__init__(
            ControlPointOperation.OpCode.MODIFY_SOURCE,
            struct.pack("<BBH", source_id, pa_sync, pa_interval)
            + encode_subgroups(subgroups),
        )
        self.source_id = source_id
        self.pa_sync = pa_sync
        self.pa_interval = pa_interval
        self.subgroups = list(subgroups)


class SetBroadcastCodeOperation(ControlPointOperation):
    @classmethod
    def from_parameters(cls, parameters: bytes) -> SetBroadcastCodeOperation:
        instance = cls.__new__(cls)
        instance.op_code = ControlPointOperation.OpCode.SET_BROADCAST_CODE
        instance.parameters = parameters
        instance.source_id = parameters[0]
        instance.broadcast_code = parameters[1:17]
        return instance

    def __init__(
        self,
        source_id: int,
        broadcast_code: bytes,
    ) -> None:
        super().__init__(
            ControlPointOperation.OpCode.SET_BROADCAST_CODE,
            bytes([source_id]) + broadcast_code,
        )
        self.source_id = source_id
        self.broadcast_code = broadcast_code

        if len(self.broadcast_code) != 16:
            raise core.InvalidArgumentError("broadcast_code must be 16 bytes")


class RemoveSourceOperation(ControlPointOperation):
    @classmethod
    def from_parameters(cls, parameters: bytes) -> RemoveSourceOperation:
        instance = cls.__new__(cls)
        instance.op_code = ControlPointOperation.OpCode.REMOVE_SOURCE
        instance.parameters = parameters
        instance.source_id = parameters[0]
        return instance

    def __init__(self, source_id: int) -> None:
        super().__init__(ControlPointOperation.OpCode.REMOVE_SOURCE, bytes([source_id]))
        self.source_id = source_id


@dataclasses.dataclass
class BroadcastReceiveState:
    class PeriodicAdvertisingSyncState(utils.OpenIntEnum):
        NOT_SYNCHRONIZED_TO_PA = 0x00
        SYNCINFO_REQUEST = 0x01
        SYNCHRONIZED_TO_PA = 0x02
        FAILED_TO_SYNCHRONIZE_TO_PA = 0x03
        NO_PAST = 0x04

    class BigEncryption(utils.OpenIntEnum):
        NOT_ENCRYPTED = 0x00
        BROADCAST_CODE_REQUIRED = 0x01
        DECRYPTING = 0x02
        BAD_CODE = 0x03

    source_id: int
    source_address: hci.Address
    source_adv_sid: int
    broadcast_id: int
    pa_sync_state: PeriodicAdvertisingSyncState
    big_encryption: BigEncryption
    bad_code: bytes
    subgroups: list[SubgroupInfo]

    @classmethod
    def from_bytes(cls, data: bytes) -> BroadcastReceiveState:
        source_id = data[0]
        _, source_address = hci.Address.parse_address_preceded_by_type(data, 2)
        source_adv_sid = data[8]
        broadcast_id = int.from_bytes(data[9:12], "little")
        pa_sync_state = cls.PeriodicAdvertisingSyncState(data[12])
        big_encryption = cls.BigEncryption(data[13])
        if big_encryption == cls.BigEncryption.BAD_CODE:
            bad_code = data[14:30]
            subgroups = decode_subgroups(data[30:])
        else:
            bad_code = b""
            subgroups = decode_subgroups(data[14:])

        return cls(
            source_id,
            source_address,
            source_adv_sid,
            broadcast_id,
            pa_sync_state,
            big_encryption,
            bad_code,
            subgroups,
        )

    def __bytes__(self) -> bytes:
        return (
            struct.pack(
                "<BB6sB3sBB",
                self.source_id,
                self.source_address.address_type,
                bytes(self.source_address),
                self.source_adv_sid,
                self.broadcast_id.to_bytes(3, "little"),
                self.pa_sync_state,
                self.big_encryption,
            )
            + self.bad_code
            + encode_subgroups(self.subgroups)
        )


# -----------------------------------------------------------------------------
class BroadcastAudioScanService(gatt.TemplateService):
    UUID = gatt.GATT_BROADCAST_AUDIO_SCAN_SERVICE

    def __init__(self):
        self.broadcast_audio_scan_control_point_characteristic = gatt.Characteristic(
            gatt.GATT_BROADCAST_AUDIO_SCAN_CONTROL_POINT_CHARACTERISTIC,
            gatt.Characteristic.Properties.WRITE
            | gatt.Characteristic.Properties.WRITE_WITHOUT_RESPONSE,
            gatt.Characteristic.WRITEABLE,
            gatt.CharacteristicValue(
                write=self.on_broadcast_audio_scan_control_point_write
            ),
        )

        self.broadcast_receive_state_characteristic = gatt.Characteristic(
            gatt.GATT_BROADCAST_RECEIVE_STATE_CHARACTERISTIC,
            gatt.Characteristic.Properties.READ | gatt.Characteristic.Properties.NOTIFY,
            gatt.Characteristic.Permissions.READABLE
            | gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION,
            b"12",  # TEST
        )

        super().__init__(
            [
                self.broadcast_audio_scan_control_point_characteristic,
                self.broadcast_receive_state_characteristic,
            ]
        )

    def on_broadcast_audio_scan_control_point_write(
        self, connection: device.Connection, value: bytes
    ) -> None:
        pass


# -----------------------------------------------------------------------------
class BroadcastAudioScanServiceProxy(gatt_client.ProfileServiceProxy):
    SERVICE_CLASS = BroadcastAudioScanService

    broadcast_audio_scan_control_point: gatt_client.CharacteristicProxy[bytes]
    broadcast_receive_states: list[
        gatt_client.CharacteristicProxy[BroadcastReceiveState | None]
    ]

    def __init__(self, service_proxy: gatt_client.ServiceProxy):
        self.service_proxy = service_proxy

        self.broadcast_audio_scan_control_point = (
            service_proxy.get_required_characteristic_by_uuid(
                gatt.GATT_BROADCAST_AUDIO_SCAN_CONTROL_POINT_CHARACTERISTIC
            )
        )

        self.broadcast_receive_states = [
            gatt_adapters.DelegatedCharacteristicProxyAdapter(
                characteristic,
                decode=lambda x: BroadcastReceiveState.from_bytes(x) if x else None,
            )
            for characteristic in service_proxy.get_characteristics_by_uuid(
                gatt.GATT_BROADCAST_RECEIVE_STATE_CHARACTERISTIC
            )
        ]

    async def send_control_point_operation(
        self, operation: ControlPointOperation
    ) -> None:
        await self.broadcast_audio_scan_control_point.write_value(
            bytes(operation), with_response=True
        )

    async def remote_scan_started(self) -> None:
        await self.send_control_point_operation(RemoteScanStartedOperation())

    async def remote_scan_stopped(self) -> None:
        await self.send_control_point_operation(RemoteScanStoppedOperation())

    async def add_source(
        self,
        advertiser_address: hci.Address,
        advertising_sid: int,
        broadcast_id: int,
        pa_sync: PeriodicAdvertisingSyncParams,
        pa_interval: int,
        subgroups: Sequence[SubgroupInfo],
    ) -> None:
        await self.send_control_point_operation(
            AddSourceOperation(
                advertiser_address,
                advertising_sid,
                broadcast_id,
                pa_sync,
                pa_interval,
                subgroups,
            )
        )

    async def modify_source(
        self,
        source_id: int,
        pa_sync: PeriodicAdvertisingSyncParams,
        pa_interval: int,
        subgroups: Sequence[SubgroupInfo],
    ) -> None:
        await self.send_control_point_operation(
            ModifySourceOperation(
                source_id,
                pa_sync,
                pa_interval,
                subgroups,
            )
        )

    async def remove_source(self, source_id: int) -> None:
        await self.send_control_point_operation(RemoveSourceOperation(source_id))
